"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from concurrent.futures.thread import ThreadPoolExecutor
import datetime
import logging
import os
import random
import uuid
import time
import asyncio
from logging import Logger

from tornado import gen, locks, ioloop
from inmanta import env, const
from inmanta import protocol
from inmanta.agent import handler
from inmanta.loader import CodeLoader
from inmanta.protocol import SessionEndpoint, methods
from inmanta.resources import Resource, Id
from tornado.concurrent import Future
from inmanta.agent.cache import AgentCache
from inmanta.agent import config as cfg
from inmanta.agent.reporting import collect_report
from typing import Tuple, Optional, Generator, Any, Dict, List, TYPE_CHECKING
from inmanta.agent.handler import ResourceHandler
from inmanta.types import NoneGen

LOGGER = logging.getLogger(__name__)
GET_RESOURCE_BACKOFF = 5


class ResourceActionResult(object):

    def __init__(self, success: bool, receive_events: bool, cancel: bool) -> None:
        self.success = success
        self.receive_events = receive_events
        self.cancel = cancel

    def __add__(self, other: "ResourceActionResult") -> "ResourceActionResult":
        return ResourceActionResult(self.success and other.success,
                                    self.receive_events or other.receive_events,
                                    self.cancel or other.cancel)

    def __str__(self) -> str:
        return "%r %r %r" % (self.success, self.receive_events, self.cancel)


# https://mypy.readthedocs.io/en/latest/common_issues.html#using-classes-that-are-generic-in-stubs-but-not-at-runtime
if TYPE_CHECKING:
    ResourceActionResultFuture = asyncio.Future[ResourceActionResult]
else:
    ResourceActionResultFuture = asyncio.Future


class ResourceAction(object):

    resource: Resource
    resource_id: Id
    future: ResourceActionResultFuture

    def __init__(self, scheduler: "ResourceScheduler", resource: Resource, gid: uuid.UUID, reason: str) -> None:
        """
            :param gid A unique identifier to identify a deploy. This is local to this agent.
        """
        self.scheduler: "ResourceScheduler" = scheduler
        self.resource: Resource = resource
        if self.resource is not None:
            self.resource_id: Id = resource.id
        self.future: ResourceActionResultFuture = Future()
        self.running: bool = False
        self.gid: uuid.UUID = gid
        self.status: Optional[const.ResourceState] = None
        self.change: Optional[const.Change] = None
        # resourceid -> attribute -> {current: , desired:}
        self.changes: Optional[Dict[str, Dict[str, Dict[str, str]]]] = None
        self.undeployable: Optional[const.ResourceState] = None
        self.reason: str = reason
        self.logger: Logger = self.scheduler.logger

    def is_running(self) -> bool:
        return self.running

    def is_done(self) -> bool:
        return self.future.done()

    def cancel(self) -> None:
        if not self.is_running() and not self.is_done():
            LOGGER.info("Cancelled deploy of %s %s", self.gid, self.resource)
            self.future.set_result(ResourceActionResult(False, False, True))

    @gen.coroutine
    def send_in_progress(self,
                         action_id: uuid.UUID,
                         start: float,
                         status: const.ResourceState = const.ResourceState.deploying) -> NoneGen:
        yield self.scheduler.get_client().resource_action_update(tid=self.scheduler._env_id,
                                                                 resource_ids=[str(self.resource.id)],
                                                                 action_id=action_id,
                                                                 action=const.ResourceAction.deploy,
                                                                 started=start,
                                                                 status=status)

    @gen.coroutine
    def _execute(self, ctx: handler.HandlerContext, events: dict, cache: AgentCache, start: float,
                 event_only: bool=False) -> Generator[Any, Any, Tuple[bool, bool]]:
        """
            :param ctx The context to use during execution of this deploy
            :param events Possible events that are available for this resource
            :param cache The cache instance to use
            :param event_only: don't execute, only do event propagation
            :param state: start time
            :return (success, send_event) Return whether the execution was successful and whether a change event should be sent
                                          to provides of this resource.
        """
        ctx.debug("Start deploy %(deploy_id)s of resource %(resource_id)s",
                  deploy_id=self.gid, resource_id=self.resource_id)
        provider: Optional[ResourceHandler] = None

        if not event_only:
            yield self.send_in_progress(ctx.action_id, start)
        else:
            yield self.send_in_progress(ctx.action_id, start, status=const.ResourceState.processing_events)

        # setup provider
        try:
            provider = yield self.scheduler.agent.get_provider(self.resource)
        except Exception:
            if provider is not None:
                provider.close()

            cache.close_version(self.resource.id.version)
            ctx.set_status(const.ResourceState.unavailable)
            ctx.exception("Unable to find a handler for %(resource_id)s", resource_id=str(self.resource.id))
            return False, False

        success = True
        # no events by default, i.e. don't send events if you are not executed
        send_event = False

        # main execution
        if not event_only:
            send_event = (hasattr(self.resource, "send_event") and self.resource.send_event)

            try:
                yield self.scheduler.agent.thread_pool.submit(provider.execute, ctx, self.resource)
            except Exception as e:
                ctx.set_status(const.ResourceState.failed)
                ctx.exception("An error occurred during deployment of %(resource_id)s (exception: %(exception)s",
                              resource_id=self.resource.id, exception=repr(e))

            if ctx.status is not const.ResourceState.deployed:
                success = False

        # event processing
        if len(events) > 0 and provider.can_process_events():
            if not event_only:
                yield self.send_in_progress(ctx.action_id, start, status=const.ResourceState.processing_events)
            try:
                ctx.info("Sending events to %(resource_id)s because of modified dependencies",
                         resource_id=str(self.resource.id))
                yield self.scheduler.agent.thread_pool.submit(provider.process_events, ctx, self.resource, events)
            except Exception:
                ctx.exception("Could not send events for %(resource_id)s",
                              resource_id=str(self.resource.id),
                              events=str(events))

        provider.close()
        cache.close_version(self.resource_id.version)

        return success, send_event

    def skipped_because(self, results):
        return [resource.resource_id.resource_str()
                for resource, result in zip(self.dependencies, results) if not result.success]

    @gen.coroutine
    def execute(self, dummy: "ResourceAction", generation: "Dict[str, ResourceAction]", cache: AgentCache) -> NoneGen:
        self.logger.log(const.LogLevel.TRACE.value, "Entering %s %s", self.gid, self.resource)
        cache.open_version(self.resource.id.version)

        self.dependencies = [generation[x.resource_str()] for x in self.resource.requires]
        waiters = [x.future for x in self.dependencies]
        waiters.append(dummy.future)
        results = yield waiters

        with (yield self.scheduler.ratelimiter.acquire()):
            start = datetime.datetime.now()
            ctx = handler.HandlerContext(self.resource, logger=self.logger)

            ctx.debug(
                "Start run for resource %(resource)s because %(reason)s",
                resource=str(self.resource.id),
                deploy_id=self.gid,
                agent=self.scheduler.agent.name,
                reason=self.reason
            )

            self.running = True
            if self.is_done():
                # Action is cancelled
                LOGGER.log(const.LogLevel.TRACE.value, "%s %s is no longer active" % (self.gid, self.resource))
                self.running = False
                if self.undeployable is not None:
                    # don't overwrite undeployable
                    ctx.set_status(self.undeployable)
                else:
                    ctx.set_status(const.ResourceState.cancelled)
                return

            result = sum(results, ResourceActionResult(True, False, False))

            if result.cancel:
                # self.running will be set to false when self.cancel is called
                # Only happens when global cancel has not cancelled us but our predecessors have already been cancelled
                if self.undeployable is not None:
                    # don't overwrite undeployable
                    ctx.set_status(self.undeployable)
                else:
                    ctx.set_status(const.ResourceState.cancelled)
                return

            if result.receive_events:
                received_events = {x.resource_id: dict(status=x.status, change=x.change,
                                                       changes=x.changes.get(str(x.resource_id), {}))
                                   for x in self.dependencies}
            else:
                received_events = {}

            if self.undeployable is not None:
                ctx.set_status(self.undeployable)
                success = False
                send_event = False
            elif not result.success:
                ctx.set_status(const.ResourceState.skipped)
                ctx.info(
                    "Resource %(resource)s skipped due to failed dependency %(failed)s",
                    resource=str(self.resource.id),
                    failed=self.skipped_because(results)
                )
                success = False
                send_event = False
                yield self._execute(ctx=ctx, events=received_events, cache=cache, event_only=True, start=start)
            else:
                success, send_event = yield self._execute(ctx=ctx, events=received_events, cache=cache, start=start)

            ctx.debug(
                "End run for resource %(resource)s in deploy %(deploy_id)s", resource=str(self.resource.id), deploy_id=self.gid
            )

            end = datetime.datetime.now()
            changes = {str(self.resource.id): ctx.changes}
            result = yield self.scheduler.get_client().resource_action_update(tid=self.scheduler._env_id,
                                                                              resource_ids=[str(self.resource.id)],
                                                                              action_id=ctx.action_id,
                                                                              action=const.ResourceAction.deploy,
                                                                              started=start, finished=end, status=ctx.status,
                                                                              changes=changes,
                                                                              messages=ctx.logs, change=ctx.change,
                                                                              send_events=send_event)
            if result.code != 200:
                LOGGER.error("Resource status update failed %s", result.result)

            self.status = ctx.status
            self.change = ctx.change
            self.changes = changes
            self.future.set_result(ResourceActionResult(success, send_event, False))
            self.running = False

    def __str__(self) -> str:
        if self.resource is None:
            return "DUMMY"

        status = ""
        if self.is_done():
            status = " Done"
        elif self.is_running():
            status = " Running"

        return self.resource.id.resource_str() + status

    def long_string(self) -> str:
        return "%s awaits %s" % (self.resource.id.resource_str(), " ".join([str(aw) for aw in self.dependencies]))


class RemoteResourceAction(ResourceAction):

    def __init__(self, scheduler: "ResourceScheduler", resource_id: Id, gid: uuid.UUID, reason: str):
        super(RemoteResourceAction, self).__init__(scheduler, None, gid, reason)
        self.resource_id = resource_id

    @gen.coroutine
    def execute(self, dummy: "ResourceAction", generation: "Dict[str, ResourceAction]", cache: AgentCache) -> NoneGen:
        yield dummy.future
        try:
            result = yield self.scheduler.get_client().get_resource(self.scheduler.agent._env_id, str(self.resource_id),
                                                                    logs=True, log_action=const.ResourceAction.deploy,
                                                                    log_limit=1)
            if result.code != 200:
                LOGGER.error("Failed to get the status for remote resource %s (%s)", str(self.resource_id),
                             result.result)

            status = const.ResourceState[result.result["resource"]["status"]]
            if status in const.TRANSIENT_STATES or self.future.done():
                # wait for event
                pass
            else:
                if status == const.ResourceState.deployed:
                    success = True
                else:
                    success = False

                send_event = False
                if "logs" in result.result and len(result.result["logs"]) > 0:
                    log = result.result["logs"][0]

                    if "change" in log:
                        self.change = const.Change[log["change"]]
                    else:
                        self.change = const.Change.nochange

                    if "changes" in log and str(self.resource_id) in log["changes"]:
                        self.changes = log["changes"]
                    else:
                        self.changes = {}
                    self.status = status
                    send_event = log["send_event"]

                self.future.set_result(ResourceActionResult(success, send_event, False))

            self.running = False
        except Exception:
            LOGGER.exception("could not get status for remote resource")

    def notify(self,
               send_events: bool,
               status: const.ResourceState,
               change: const.Change,
               changes: Dict[str, Dict[str, Dict[str, str]]]) -> None:
        if not self.future.done():
            self.status = status
            self.change = change
            self.changes = changes
            self.future.set_result(ResourceActionResult(True, send_events, False))


class ResourceScheduler(object):
    """Class responsible for managing sequencing of actions performed by the agent.

    State of the last run is not removed after the run but remains.

    By not removing e.g. the generation,
    1 - the class is always in a valid state
    2 - we don't need to figure out exactly when a run is done
    """

    def __init__(self,
                 agent: "AgentInstance",
                 env_id: uuid.UUID,
                 name: str,
                 cache: AgentCache,
                 ratelimiter: locks.Semaphore) -> None:
        self.generation: Dict[str, ResourceAction] = {}
        self.cad: Dict[str, RemoteResourceAction] = {}
        self._env_id = env_id
        self.agent = agent
        self.cache = cache
        self.name = name
        self.ratelimiter = ratelimiter
        self.version: int = 0
        # the reason the last run was started
        self.reason: str = ""
        # was the last run a repair run?
        self.is_repair: bool = False
        # if this value is not None, a new repair run will be started after the current run is done
        # this field is both flag and value, to ensure consistency
        self._resume_reason: Optional[str] = None
        self.logger: Logger = agent.logger

    def get_scheduled_resource_actions(self) -> List[ResourceAction]:
        return list(self.generation.values())

    def finished(self) -> bool:
        for resource_action in self.generation.values():
            if not resource_action.is_done():
                return False
        return True

    def is_normal_deploy_running(self) -> bool:
        return not self.finished() and not self.is_repair

    def reload(self, resources, undeployable={}, reason: str="RELOAD", is_repair=False) -> None:
        """
        Schedule a new set of resources for execution.

        **This method should only be called under critical_ratelimiter lock!**
        """

        # First determined if we should start and if the current run should be resumed
        if not self.finished():
            # we are still running
            if self.is_repair:
                # now running repair
                if is_repair:
                    # repair restarts repair
                    self.logger.info("Terminating run '%s' for '%s'", self.reason, reason)
                else:
                    # increment interrupts repair
                    self.logger.info("Interrupting run '%s' for '%s'", self.reason, reason)
                    self._resume_reason = "Restarting run '%s', interrupted for '%s'" % (self.reason,
                                                                                         reason)
            else:
                # now running increment
                if is_repair:
                    # repair is delayed
                    self.logger.info("Deferring run '%s' for '%s'", reason, self.reason)
                    self._resume_reason = reason
                    return
                else:
                    # increment overrules increment
                    self.logger.info("Terminating run '%s' for '%s'", self.reason, reason)
            # cancel old run
            for ra in self.generation.values():
                ra.cancel()

        # start new run
        self.reason = reason
        self.is_repair = is_repair
        version = resources[0].id.get_version
        self.version = version
        gid = uuid.uuid4()
        self.logger.info("Running %s for reason: %s" % (gid, reason))

        # re-generate generation
        self.generation = {r.id.resource_str(): ResourceAction(self, r, gid, reason) for r in resources}

        # mark undeployable
        for key, res in self.generation.items():
            vid = str(res.resource.id)
            if vid in undeployable:
                self.generation[key].undeployable = undeployable[vid]

        # hook up Cross Agent Dependencies
        cross_agent_dependencies = [q for r in resources for q in r.requires if q.get_agent_name() != self.name]
        for cad in cross_agent_dependencies:
            ra = RemoteResourceAction(self, cad, gid, reason)
            self.cad[str(cad)] = ra
            self.generation[cad.resource_str()] = ra

        # Create dummy to give start signal
        dummy = ResourceAction(self, None, gid, reason)
        # Dispatch all actions
        # Will block on dependencies and dummy
        for r in self.generation.values():
            self.agent.add_future(r.execute(dummy, self.generation, self.cache))

        # Listen for completion
        self.agent.add_future(self.mark_deployment_as_finished(self.generation.values(), reason, gid))

        # Start running
        dummy.future.set_result(ResourceActionResult(True, False, False))

    @gen.coroutine
    def mark_deployment_as_finished(self, resource_actions, reason, gid):
        futures = [resource_action.future for resource_action in resource_actions]
        yield futures  # Wait until deployment finishes
        with (yield self.agent.critical_ratelimiter.acquire()):
            if not self.finished():
                return
            if self._resume_reason is not None:
                self.logger.info("Resuming run '%s'", self._resume_reason)
                self.agent.add_future(self.agent.get_latest_version_for_agent(reason=self._resume_reason,
                                                                              incremental_deploy=False, is_repair_run=True))
                self._resume_reason = None

    def notify_ready(self, resourceid, send_events, state, change, changes):
        if resourceid not in self.cad:
            # received CAD notification for which no resource are waiting, so return
            return
        self.cad[resourceid].notify(send_events, state, change, changes)

    def dump(self):
        print("Waiting:")
        for r in self.generation.values():
            print(r.long_string())
        print("Ready to run:")
        for r in self.queue:
            print(r.long_string())

    def get_client(self):
        return self.agent.get_client()


class AgentInstance(object):

    def __init__(self, process: "Agent", name: str, uri: str) -> None:
        self.process = process
        self.name = name
        self._uri = uri

        self.logger: Logger = LOGGER.getChild(self.name)

        # the lock for changing the current ongoing deployment
        self.critical_ratelimiter = locks.Semaphore(1)
        # lock for dryrun tasks
        self.dryrunlock = locks.Semaphore(1)

        # multi threading control
        # threads to setup connections
        self.provider_thread_pool: ThreadPoolExecutor = ThreadPoolExecutor(1, thread_name_prefix="ProviderPool_%s" % name)
        # threads to work
        self.thread_pool: ThreadPoolExecutor = ThreadPoolExecutor(process.poolsize, thread_name_prefix="Pool_%s" % name)
        self.ratelimiter = locks.Semaphore(process.poolsize)

        self._env_id: uuid.UUID = process._env_id

        self.sessionid = process.sessionid

        # init
        self._cache = AgentCache()
        self._nq = ResourceScheduler(self, self.process.environment, name, self._cache, ratelimiter=self.ratelimiter)
        self._time_triggered_actions = set()
        self._enabled = False

        # do regular deploys
        self._deploy_interval = cfg.agent_deploy_interval.get()
        deploy_splay_time = cfg.agent_deploy_splay_time.get()
        self._deploy_splay_value = random.randint(0, deploy_splay_time)

        # do regular repair runs
        self._repair_interval = cfg.agent_repair_interval.get()
        repair_splay_time = cfg.agent_repair_splay_time.get()
        self._repair_splay_value = random.randint(0, repair_splay_time)

        self._getting_resources = False
        self._get_resource_timeout = 0

    @gen.coroutine
    def stop(self):
        self.provider_thread_pool.shutdown(wait=False)
        self.thread_pool.shutdown(wait=False)

    @property
    def environment(self):
        return self.process.environment

    def get_client(self):
        return self.process._client

    @property
    def uri(self):
        return self._uri

    def is_enabled(self):
        return self._enabled

    def add_future(self, future):
        self.process.add_future(future)

    def unpause(self):
        if self.is_enabled():
            return 200, "already running"

        self.logger.info("Agent assuming primary role for %s", self.name)

        self._enable_time_triggers()
        self._enabled = True
        return 200, "unpaused"

    def pause(self, reason="agent lost primary role"):
        if not self.is_enabled():
            return 200, "already paused"

        self.logger.info("Agent %s stopped because %s", self.name, reason)

        self._disable_time_triggers()
        self._enabled = False
        return 200, "paused"

    def _enable_time_triggers(self):

        @gen.coroutine
        def deploy_action():
            now = datetime.datetime.now()
            yield self.get_latest_version_for_agent(
                reason="Periodic deploy started at %s" % (now.strftime(const.TIME_LOGFMT)),
                incremental_deploy=True,
                is_repair_run=False)

        @gen.coroutine
        def repair_action():
            now = datetime.datetime.now()
            yield self.get_latest_version_for_agent(
                reason="Repair run started at %s" % (now.strftime(const.TIME_LOGFMT)),
                incremental_deploy=False,
                is_repair_run=True)

        now = datetime.datetime.now()
        if self._deploy_interval > 0:
            self.logger.info("Scheduling periodic deploy with interval %d and splay %d (first run at %s)",
                             self._deploy_interval,
                             self._deploy_splay_value,
                             (now + datetime.timedelta(seconds=self._deploy_splay_value)).strftime(const.TIME_LOGFMT)
                             )
            self._enable_time_trigger(deploy_action, self._deploy_interval, self._deploy_splay_value)
        if self._repair_interval > 0:
            self.logger.info("Scheduling repair with interval %d and splay %d (first run at %s)",
                             self._repair_interval,
                             self._repair_splay_value,
                             (now + datetime.timedelta(seconds=self._repair_splay_value)).strftime(const.TIME_LOGFMT)
                             )
            self._enable_time_trigger(repair_action, self._repair_interval, self._repair_splay_value)

    def _enable_time_trigger(self, action, interval, splay):
        self.process._sched.add_action(action, interval, splay)
        self._time_triggered_actions.add(action)

    def _disable_time_triggers(self):
        for action in self._time_triggered_actions:
            self.process._sched.remove(action)
        self._time_triggered_actions.clear()

    def notify_ready(self, resourceid, send_events, state, change, changes):
        self._nq.notify_ready(resourceid, send_events, state, change, changes)

    def _can_get_resources(self):
        if self._getting_resources:
            self.logger.info("Attempting to get resource while get is in progress")
            return False
        if time.time() < self._get_resource_timeout:
            self.logger.info("Attempting to get resources during backoff %g seconds left, last download took %d seconds",
                             self._get_resource_timeout - time.time(), self._get_resource_duration)
            return False
        return True

    @gen.coroutine
    def get_provider(self, resource: Resource) -> Generator[Any, Any, ResourceHandler]:
        provider = yield self.provider_thread_pool.submit(handler.Commander.get_provider, self._cache, self, resource)
        provider.set_cache(self._cache)
        return provider

    @gen.coroutine
    def get_latest_version_for_agent(self, reason="Unknown", incremental_deploy=False, is_repair_run=False):
        """
            Get the latest version for the given agent (this is also how we are notified)

            :param reason: the reason this deploy was started
        """
        if not self._can_get_resources():
            self.logger.warning("%s aborted by rate limiter", reason)
            return
        with (yield self.critical_ratelimiter.acquire()):
            if not self._can_get_resources():
                self.logger.warning("%s aborted by rate limiter", reason)
                return

            self.logger.debug("Getting latest resources for %s", reason)
            self._getting_resources = True
            start = time.time()
            try:
                result = yield self.get_client().get_resources_for_agent(tid=self._env_id, agent=self.name,
                                                                         incremental_deploy=incremental_deploy)
            finally:
                self._getting_resources = False
            end = time.time()
            self._get_resource_duration = end - start
            self._get_resource_timeout = GET_RESOURCE_BACKOFF * self._get_resource_duration + end
            if result.code == 404:
                self.logger.info("No released configuration model version available for %s", reason)
            elif result.code == 409:
                self.logger.warning("We are not currently primary during %s: %s", reason, result.result)
            elif result.code != 200:
                self.logger.warning("Got an error while pulling resources for %s. %s",
                                    reason,
                                    result.result)

            else:
                restypes = set([res["resource_type"] for res in result.result["resources"]])
                resources = []
                yield self.process._ensure_code(self._env_id, result.result["version"], restypes)
                try:
                    undeployable = {}
                    for res in result.result["resources"]:
                        state = const.ResourceState[res["status"]]
                        if state in const.UNDEPLOYABLE_STATES:
                            undeployable[res["id"]] = state

                        data = res["attributes"]
                        data["id"] = res["id"]
                        resource = Resource.deserialize(data)
                        resources.append(resource)
                        self.logger.debug("Received update for %s", resource.id)
                except TypeError:
                    self.logger.exception("Failed to receive update for %s", reason)

                self.logger.debug("Pulled %d resources because %s", len(resources), reason)

                if len(resources) > 0:
                    self._nq.reload(resources, undeployable, reason=reason, is_repair=is_repair_run)

    @gen.coroutine
    def dryrun(self, dry_run_id, version):
        self.add_future(self.do_run_dryrun(version, dry_run_id))
        return 200

    @gen.coroutine
    def do_run_dryrun(self, version, dry_run_id):
        with (yield self.dryrunlock.acquire()):
            with (yield self.ratelimiter.acquire()):
                result = yield self.get_client().get_resources_for_agent(tid=self._env_id, agent=self.name, version=version)
                if result.code == 404:
                    self.logger.warning("Version %s does not exist, can not run dryrun", version)
                    return

                elif result.code != 200:
                    self.logger.warning("Got an error while pulling resources and version %s", version)
                    return

                resources = result.result["resources"]
                restypes = set([res["resource_type"] for res in resources])

                # TODO: handle different versions for dryrun and deploy!
                yield self.process._ensure_code(self._env_id, version, restypes)

                self._cache.open_version(version)

                for res in resources:
                    ctx = handler.HandlerContext(res, True)
                    started = datetime.datetime.now()
                    provider = None
                    try:
                        if const.ResourceState[res["status"]] in const.UNDEPLOYABLE_STATES:
                            ctx.exception("Skipping %(resource_id)s because in undeployable state %(status)s",
                                          resource_id=res["id"], status=res["status"])
                            yield self.get_client().dryrun_update(tid=self._env_id, id=dry_run_id, resource=res["id"],
                                                                  changes={})
                            continue

                        data = res["attributes"]
                        data["id"] = res["id"]
                        resource = Resource.deserialize(data)
                        self.logger.debug("Running dryrun for %s", resource.id)

                        try:
                            provider = yield self.get_provider(resource)
                        except Exception as e:
                            ctx.exception("Unable to find a handler for %(resource_id)s (exception: %(exception)s",
                                          resource_id=str(resource.id), exception=str(e))
                            yield self.get_client().dryrun_update(tid=self._env_id, id=dry_run_id, resource=res["id"],
                                                                  changes={"handler": {"current": "FAILED",
                                                                                       "desired": "Unable to find a handler"}})
                        else:
                            try:
                                yield self.thread_pool.submit(provider.execute, ctx, resource, dry_run=True)
                                changes = ctx.changes
                                if changes is None:
                                    changes = {}
                                if(ctx.status == const.ResourceState.failed):
                                    changes["handler"] = {"current": "FAILED", "desired": "Handler failed"}
                                yield self.get_client().dryrun_update(tid=self._env_id, id=dry_run_id, resource=res["id"],
                                                                      changes=changes)
                            except Exception as e:
                                ctx.exception("Exception during dryrun for %(resource_id)s (exception: %(exception)s",
                                              resource_id=str(resource.id), exception=str(e))
                                changes = ctx.changes
                                if changes is None:
                                    changes = {}
                                changes["handler"] = {"current": "FAILED", "desired": "Handler failed"}
                                yield self.get_client().dryrun_update(tid=self._env_id, id=dry_run_id, resource=res["id"],
                                                                      changes=changes)

                    except Exception:
                        ctx.exception("Unable to process resource for dryrun.")
                        changes = {}
                        changes["handler"] = {"current": "FAILED", "desired": "Resource Deserialization Failed"}
                        yield self.get_client().dryrun_update(tid=self._env_id, id=dry_run_id, resource=res["id"],
                                                              changes=changes)
                    finally:
                        if provider is not None:
                            provider.close()

                        finished = datetime.datetime.now()
                        yield self.get_client().resource_action_update(tid=self._env_id, resource_ids=[res["id"]],
                                                                       action_id=ctx.action_id,
                                                                       action=const.ResourceAction.dryrun,
                                                                       started=started,
                                                                       finished=finished,
                                                                       messages=ctx.logs,
                                                                       status=const.ResourceState.dry)

                self._cache.close_version(version)

    @gen.coroutine
    def get_facts(self, resource):
        with (yield self.ratelimiter.acquire()):
            yield self.process._ensure_code(self._env_id, resource["model"], [resource["resource_type"]])
            ctx = handler.HandlerContext(resource)
            started = datetime.datetime.now()
            provider = None
            try:
                data = resource["attributes"]
                data["id"] = resource["id"]
                resource_obj = Resource.deserialize(data)

                version = resource_obj.id.version
                try:
                    self._cache.open_version(version)
                    provider = yield self.get_provider(resource_obj)
                    result = yield self.thread_pool.submit(provider.check_facts, ctx, resource_obj)
                    parameters = [{"id": name, "value": value, "resource_id": resource_obj.id.resource_str(), "source": "fact"}
                                  for name, value in result.items()]
                    yield self.get_client().set_parameters(tid=self._env_id, parameters=parameters)
                    finished = datetime.datetime.now()
                    yield self.get_client().resource_action_update(tid=self._env_id,
                                                                   resource_ids=[resource_obj.id.resource_str()],
                                                                   action_id=ctx.action_id,
                                                                   action=const.ResourceAction.getfact,
                                                                   started=started,
                                                                   finished=finished,
                                                                   messages=ctx.logs)

                except Exception:
                    self.logger.exception("Unable to retrieve fact")
                finally:
                    self._cache.close_version(version)

            except Exception:
                self.logger.exception("Unable to find a handler for %s", resource["id"])
                return 500
            finally:
                if provider is not None:
                    provider.close()
            return 200


class Agent(SessionEndpoint):
    """
        An agent to enact changes upon resources. This agent listens to the
        message bus for changes.
    """

    def __init__(self, hostname=None, agent_map=None, code_loader=True, environment=None, poolsize=1,
                 cricital_pool_size=5):
        super().__init__("agent", timeout=cfg.server_timeout.get(), reconnect_delay=cfg.agent_reconnect_delay.get())

        self.poolsize = poolsize
        self.ratelimiter = locks.Semaphore(poolsize)
        self.critical_ratelimiter = locks.Semaphore(cricital_pool_size)
        self.thread_pool = ThreadPoolExecutor(poolsize, thread_name_prefix="mainpool")

        if agent_map is None:
            agent_map = cfg.agent_map.get()

        self.agent_map = agent_map
        self._storage = self.check_storage()

        if environment is None:
            environment = cfg.environment.get()
            if environment is None:
                raise Exception("The agent requires an environment to be set.")
        self.set_environment(environment)

        self._instances = {}

        if code_loader:
            self._env = env.VirtualEnv(self._storage["env"])
            self._env.use_virtual_env()
            self._loader = CodeLoader(self._storage["code"])
        else:
            self._loader = None

        if hostname is not None:
            self.add_end_point_name(hostname)

        else:
            # load agent names from the config file
            agent_names = cfg.agent_names.get()
            if agent_names is not None:
                names = [x.strip() for x in agent_names.split(",")]
                for name in names:
                    if "$" in name:
                        name = name.replace("$node-name", self.node_name)

                    self.add_end_point_name(name)

        # cache reference to THIS ioloop for handlers to push requests on it
        # defer to start, just to be sure
        self._io_loop = None

    @gen.coroutine
    def stop(self):
        yield super(Agent, self).stop()
        self.thread_pool.shutdown(wait=False)
        for instance in self._instances.values():
            yield instance.stop()

    @gen.coroutine
    def start(self):
        # cache reference to THIS ioloop for handlers to push requests on it
        self._io_loop = ioloop.IOLoop.current()
        yield super(Agent, self).start()

    def add_end_point_name(self, name):
        SessionEndpoint.add_end_point_name(self, name)

        hostname = "local:"
        if name in self.agent_map:
            hostname = self.agent_map[name]

        self._instances[name] = AgentInstance(self, name, hostname)

    def unpause(self, name):
        if name not in self._instances:
            return 404, "No such agent"

        return self._instances[name].unpause()

    def pause(self, name):
        if name not in self._instances:
            return 404, "No such agent"

        return self._instances[name].pause()

    @protocol.handle(methods.set_state)
    @gen.coroutine
    def set_state(self, agent, enabled):
        if enabled:
            return self.unpause(agent)
        else:
            return self.pause(agent)

    @gen.coroutine
    def on_reconnect(self) -> NoneGen:
        for name in self._instances.keys():
            result = yield self._client.get_state(tid=self._env_id, sid=self.sessionid, agent=name)
            if result.code == 200:
                state = result.result
                if "enabled" in state and isinstance(state["enabled"], bool):
                    self.set_state(name, state["enabled"])
                else:
                    LOGGER.warning("Server reported invalid state %s" % (repr(state)))
            else:
                LOGGER.warning("could not get state from the server")

    @gen.coroutine
    def on_disconnect(self) -> NoneGen:
        LOGGER.warning("Connection to server lost, taking agents offline")
        for agent_instance in self._instances.values():
            agent_instance.pause("Connection to server lost")

    @gen.coroutine
    def get_latest_version(self):
        """
            Get the latest version of managed resources for all agents
        """
        for agent in self._instances.values():
            yield agent.get_latest_version_for_agent(reason="call to get_latest_version on agent")

    @gen.coroutine
    def _ensure_code(self, environment, version, resourcetypes):
        """
            Ensure that the code for the given environment and version is loaded
        """
        if self._loader is not None:
            for rt in resourcetypes:
                result = yield self._client.get_code(environment, version, rt)

                if result.code == 200:
                    for hash_value, (path, name, content, requires) in result.result["sources"].items():
                        try:
                            LOGGER.debug("Installing handler %s for %s", rt, name)
                            yield self._install(hash_value, name, content, requires)
                            LOGGER.debug("Installed handler %s for %s", rt, name)
                        except Exception:
                            LOGGER.exception("Failed to install handler %s for %s", rt, name)

    @gen.coroutine
    def _install(self, hash_value, module_name, module_source, module_requires):
        yield self.thread_pool.submit(self._env.install_from_list, module_requires, True)
        yield self.thread_pool.submit(self._loader.deploy_version, hash_value, module_name, module_source)

    @protocol.handle(methods.trigger, env="tid", agent="id")
    @gen.coroutine
    def trigger_update(self, env, agent, incremental_deploy):
        """
            Trigger an update
        """
        if agent not in self._instances:
            return 200

        if not self._instances[agent].is_enabled():
            return 500, "Agent is not _enabled"

        LOGGER.info("Agent %s got a trigger to update in environment %s", agent, env)
        future = self._instances[agent].get_latest_version_for_agent(reason="call to trigger_update",
                                                                     incremental_deploy=incremental_deploy)
        self.add_future(future)
        return 200

    @protocol.handle(methods.resource_event, env="tid", agent="id")
    @gen.coroutine
    def resource_event(self, env, agent: str, resource: str, send_events: bool,
                       state: const.ResourceState, change: const.Change, changes: dict):
        if env != self._env_id:
            LOGGER.warning("received unexpected resource event: tid: %s, agent: %s, resource: %s, state: %s, tid unknown",
                           env, agent, resource, state)
            return 200

        if agent not in self._instances:
            LOGGER.warning("received unexpected resource event: tid: %s, agent: %s, resource: %s, state: %s, agent unknown",
                           env, agent, resource, state)
            return 200

        LOGGER.debug("Agent %s got a resource event: tid: %s, agent: %s, resource: %s, state: %s",
                     agent, env, agent, resource, state)
        self._instances[agent].notify_ready(resource, send_events, state, change, changes)

        return 200

    @protocol.handle(methods.do_dryrun, env="tid", dry_run_id="id")
    @gen.coroutine
    def run_dryrun(self, env, dry_run_id, agent, version):
        """
           Run a dryrun of the given version
        """
        assert env == self._env_id

        if agent not in self._instances:
            return 200

        LOGGER.info("Agent %s got a trigger to run dryrun %s for version %s in environment %s",
                    agent, dry_run_id, version, env)

        return (yield self._instances[agent].dryrun(dry_run_id, version))

    def check_storage(self):
        """
            Check if the server storage is configured and ready to use.
        """

        state_dir = cfg.state_dir.get()

        if not os.path.exists(state_dir):
            os.mkdir(state_dir)

        agent_state_dir = os.path.join(state_dir, "agent")

        if not os.path.exists(agent_state_dir):
            os.mkdir(agent_state_dir)

        dir_map = {"agent": agent_state_dir}

        code_dir = os.path.join(agent_state_dir, "code")
        dir_map["code"] = code_dir
        if not os.path.exists(code_dir):
            os.mkdir(code_dir)

        env_dir = os.path.join(agent_state_dir, "env")
        dir_map["env"] = env_dir
        if not os.path.exists(env_dir):
            os.mkdir(env_dir)

        return dir_map

    @protocol.handle(methods.get_parameter, env="tid")
    @gen.coroutine
    def get_facts(self, env, agent, resource):
        if agent not in self._instances:
            return 200

        return (yield self._instances[agent].get_facts(resource))

    @protocol.handle(methods.get_status)
    @gen.coroutine
    def get_status(self):
        return 200, collect_report(self)



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio
import logging
import time
import uuid
from collections import defaultdict, namedtuple
from threading import Condition

from pytest import fixture

from inmanta import const, data
from inmanta.agent.agent import Agent
from inmanta.agent.handler import (
    provider,
    ResourceHandler,
    SkipResource,
    HandlerContext,
    CRUDHandler,
    ResourcePurged,
)
from inmanta.resources import (
    resource,
    Resource,
    PurgeableResource,
    IgnoreResourceException,
)
from inmanta.server import SLICE_AGENT_MANAGER
from utils import retry_limited

logger = logging.getLogger("inmanta.test.server_agent")


async def get_agent(server, environment, *endpoints, hostname="nodes1"):
    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)
    agent = Agent(
        hostname=hostname,
        environment=environment,
        agent_map={agent: "localhost" for agent in endpoints},
        code_loader=False,
    )
    for agentname in endpoints:
        agent.add_end_point_name(agentname)
    await agent.start()
    await retry_limited(lambda: len(agentmanager.sessions) == 1, 10)
    return agent


def get_resource(version, key="key1", agent="agent1", value="value1"):
    return {
        "key": key,
        "value": value,
        "id": f"test::Resource[{agent},key={key}],v=%d" % version,
        "send_event": False,
        "purged": False,
        "requires": [],
    }


async def _deploy_resources(
    client, environment, resources, version, push, agent_trigger_method=None
):
    result = await client.put_version(
        tid=environment,
        version=version,
        resources=resources,
        unknowns=[],
        version_info={},
    )
    assert result.code == 200

    # do a deploy
    result = await client.release_version(
        environment, version, push, agent_trigger_method
    )
    assert result.code == 200

    assert not result.result["model"]["deployed"]
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == len(resources)

    result = await client.get_version(environment, version)
    assert result.code == 200

    return result


async def _wait_until_deployment_finishes(client, environment, version, timeout=10):
    async def is_deployment_finished():
        result = await client.get_version(environment, version)
        return result.result["model"]["total"] - result.result["model"]["done"] <= 0

    await retry_limited(is_deployment_finished, timeout)


async def _wait_for_n(client, environment, version, n, timeout=10):
    async def is_deployment_finished():
        result = await client.get_version(environment, version)
        return result.result["model"]["done"] < n

    await retry_limited(is_deployment_finished, timeout)


async def _wait_for_n_deploying(client, environment, version, n, timeout=10):
    async def in_progress():
        result = await client.get_version(environment, version)
        assert result.code == 200
        res = [
            res for res in result.result["resources"] if res["status"] == "deploying"
        ]
        return len(res) >= n

    await retry_limited(in_progress, timeout)


ResourceContainer = namedtuple(
    "ResourceContainer",
    [
        "Provider",
        "waiter",
        "wait_for_done_with_waiters",
        "wait_for_condition_with_waiters",
    ],
)


@fixture(scope="function")
def resource_container():
    @resource("test::Resource", agent="agent", id_attribute="key")
    class MyResource(Resource):
        """
            A file on a filesystem
        """

        fields = ("key", "value", "purged")

    @resource("test::Fact", agent="agent", id_attribute="key")
    class FactResource(Resource):
        """
            A file on a filesystem
        """

        fields = ("key", "value", "purged", "skip", "factvalue", "skipFact")

    @resource("test::Fail", agent="agent", id_attribute="key")
    class FailR(Resource):
        """
            A file on a filesystem
        """

        fields = ("key", "value", "purged")

    @resource("test::Wait", agent="agent", id_attribute="key")
    class WaitR(Resource):
        """
            A file on a filesystem
        """

        fields = ("key", "value", "purged")

    @resource("test::WaitEvent", agent="agent", id_attribute="key")
    class WaitER(Resource):
        """
            A file on a filesystem
        """

        fields = ("key", "value", "purged")

    @resource("test::Noprov", agent="agent", id_attribute="key")
    class NoProv(Resource):
        """
            A file on a filesystem
        """

        fields = ("key", "value", "purged")

    @resource("test::FailFast", agent="agent", id_attribute="key")
    class FailFastR(Resource):
        """
            A file on a filesystem
        """

        fields = ("key", "value", "purged")

    @resource("test::BadEvents", agent="agent", id_attribute="key")
    class BadeEventR(Resource):
        """
            A file on a filesystem
        """

        fields = ("key", "value", "purged")

    @resource("test::BadPost", agent="agent", id_attribute="key")
    class BadPostR(Resource):
        """
            A file on a filesystem
        """

        fields = ("key", "value", "purged")

    @provider("test::Resource", name="test_resource")
    class Provider(ResourceHandler):
        def check_resource(self, ctx, resource):
            self.read(resource.id.get_agent_name(), resource.key)
            assert resource.value != const.UNKNOWN_STRING
            current = resource.clone()
            current.purged = not self.isset(resource.id.get_agent_name(), resource.key)

            if not current.purged:
                current.value = self.get(resource.id.get_agent_name(), resource.key)
            else:
                current.value = None

            return current

        def do_changes(self, ctx, resource, changes):
            if self.skip(resource.id.get_agent_name(), resource.key):
                raise SkipResource()

            if self.fail(resource.id.get_agent_name(), resource.key):
                raise Exception("Failed")

            if "purged" in changes:
                self.touch(resource.id.get_agent_name(), resource.key)
                if changes["purged"]["desired"]:
                    self.delete(resource.id.get_agent_name(), resource.key)
                    ctx.set_purged()
                else:
                    self.set(resource.id.get_agent_name(), resource.key, resource.value)
                    ctx.set_created()

            elif "value" in changes:
                ctx.info(
                    "Set key '%(key)s' to value '%(value)s'",
                    key=resource.key,
                    value=resource.value,
                )
                self.touch(resource.id.get_agent_name(), resource.key)
                self.set(resource.id.get_agent_name(), resource.key, resource.value)
                ctx.set_updated()

            return changes

        def facts(self, ctx, resource):
            return {
                "length": len(self.get(resource.id.get_agent_name(), resource.key)),
                "key1": "value1",
                "key2": "value2",
            }

        def can_process_events(self) -> bool:
            return True

        def process_events(self, ctx, resource, events):
            self.__class__._EVENTS[resource.id.get_agent_name()][resource.key].append(
                events
            )
            super(Provider, self).process_events(ctx, resource, events)

        def can_reload(self) -> bool:
            return True

        def do_reload(self, ctx, resource):
            self.__class__._RELOAD_COUNT[resource.id.get_agent_name()][
                resource.key
            ] += 1

        _STATE = defaultdict(dict)
        _WRITE_COUNT = defaultdict(lambda: defaultdict(lambda: 0))
        _RELOAD_COUNT = defaultdict(lambda: defaultdict(lambda: 0))
        _READ_COUNT = defaultdict(lambda: defaultdict(lambda: 0))
        _TO_SKIP = defaultdict(lambda: defaultdict(lambda: 0))
        _TO_FAIL = defaultdict(lambda: defaultdict(lambda: 0))

        _EVENTS = defaultdict(lambda: defaultdict(lambda: []))

        @classmethod
        def set_skip(cls, agent, key, skip):
            cls._TO_SKIP[agent][key] = skip

        @classmethod
        def set_fail(cls, agent, key, failcount):
            cls._TO_FAIL[agent][key] = failcount

        @classmethod
        def skip(cls, agent, key):
            doskip = cls._TO_SKIP[agent][key]
            if doskip == 0:
                return False
            cls._TO_SKIP[agent][key] -= 1
            return True

        @classmethod
        def fail(cls, agent, key):
            doskip = cls._TO_FAIL[agent][key]
            if doskip == 0:
                return False
            cls._TO_FAIL[agent][key] -= 1
            return True

        @classmethod
        def touch(cls, agent, key):
            cls._WRITE_COUNT[agent][key] += 1

        @classmethod
        def read(cls, agent, key):
            cls._READ_COUNT[agent][key] += 1

        @classmethod
        def set(cls, agent, key, value):
            cls._STATE[agent][key] = value

        @classmethod
        def get(cls, agent, key):
            if key in cls._STATE[agent]:
                return cls._STATE[agent][key]
            return None

        @classmethod
        def isset(cls, agent, key):
            return key in cls._STATE[agent]

        @classmethod
        def delete(cls, agent, key):
            if cls.isset(agent, key):
                del cls._STATE[agent][key]

        @classmethod
        def changecount(cls, agent, key):
            return cls._WRITE_COUNT[agent][key]

        @classmethod
        def readcount(cls, agent, key):
            return cls._READ_COUNT[agent][key]

        @classmethod
        def getevents(cls, agent, key):
            return cls._EVENTS[agent][key]

        @classmethod
        def reloadcount(cls, agent, key):
            return cls._RELOAD_COUNT[agent][key]

        @classmethod
        def reset(cls):
            cls._STATE = defaultdict(dict)
            cls._EVENTS = defaultdict(lambda: defaultdict(lambda: []))
            cls._WRITE_COUNT = defaultdict(lambda: defaultdict(lambda: 0))
            cls._READ_COUNT = defaultdict(lambda: defaultdict(lambda: 0))
            cls._TO_SKIP = defaultdict(lambda: defaultdict(lambda: 0))
            cls._RELOAD_COUNT = defaultdict(lambda: defaultdict(lambda: 0))

    @provider("test::Fail", name="test_fail")
    class Fail(ResourceHandler):
        def check_resource(self, ctx, resource):
            current = resource.clone()
            current.purged = not Provider.isset(
                resource.id.get_agent_name(), resource.key
            )

            if not current.purged:
                current.value = Provider.get(resource.id.get_agent_name(), resource.key)
            else:
                current.value = None

            return current

        def do_changes(self, ctx, resource, changes):
            raise Exception()

    @provider("test::FailFast", name="test_failfast")
    class FailFast(ResourceHandler):
        def check_resource(self, ctx, resource):
            raise Exception()

    @provider("test::Fact", name="test_fact")
    class Fact(ResourceHandler):
        def check_resource(self, ctx, resource):
            current = resource.clone()
            current.purged = not Provider.isset(
                resource.id.get_agent_name(), resource.key
            )

            current.value = "that"

            return current

        def do_changes(self, ctx, resource, changes):
            if resource.skip:
                raise SkipResource("can not deploy")
            if "purged" in changes:
                if changes["purged"]["desired"]:
                    Provider.delete(resource.id.get_agent_name(), resource.key)
                    ctx.set_purged()
                else:
                    Provider.set(resource.id.get_agent_name(), resource.key, "x")
                    ctx.set_created()
            else:
                ctx.set_updated()

        def facts(self, ctx: HandlerContext, resource: Resource) -> dict:
            if not Provider.isset(resource.id.get_agent_name(), resource.key):
                return {}
            elif resource.skipFact:
                raise SkipResource("Not ready")
            return {"fact": resource.factvalue}

    @provider("test::BadEvents", name="test_bad_events")
    class BadEvents(ResourceHandler):
        def check_resource(self, ctx, resource):
            current = resource.clone()
            return current

        def do_changes(self, ctx, resource, changes):
            pass

        def can_process_events(self) -> bool:
            return True

        def process_events(self, ctx, resource, events):
            raise Exception()

    @provider("test::BadPost", name="test_bad_posts")
    class BadPost(Provider):
        def post(self, ctx, resource) -> None:
            raise Exception()

    @resource("test::AgentConfig", agent="agent", id_attribute="agentname")
    class AgentConfig(PurgeableResource):
        """
            A resource that can modify the agentmap for autostarted agents
        """

        fields = ("agentname", "uri", "autostart")

        @staticmethod
        def get_autostart(exp, obj):
            try:
                if not obj.autostart:
                    raise IgnoreResourceException()
            except Exception as e:
                # When this attribute is not set, also ignore it
                raise IgnoreResourceException() from e
            return obj.autostart

    @provider("test::AgentConfig", name="agentrest")
    class AgentConfigHandler(CRUDHandler):
        def _get_map(self) -> dict:
            def call():
                return self.get_client().get_setting(
                    tid=self._agent.environment, id=data.AUTOSTART_AGENT_MAP
                )

            value = self.run_sync(call)
            return value.result["value"]

        def _set_map(self, agent_config: dict) -> None:
            def call():
                return self.get_client().set_setting(
                    tid=self._agent.environment,
                    id=data.AUTOSTART_AGENT_MAP,
                    value=agent_config,
                )

            return self.run_sync(call)

        def read_resource(self, ctx: HandlerContext, resource: AgentConfig) -> None:
            agent_config = self._get_map()
            ctx.set("map", agent_config)

            if resource.agentname not in agent_config:
                raise ResourcePurged()

            resource.uri = agent_config[resource.agentname]

        def create_resource(self, ctx: HandlerContext, resource: AgentConfig) -> None:
            agent_config = ctx.get("map")
            agent_config[resource.agentname] = resource.uri
            self._set_map(agent_config)

        def delete_resource(self, ctx: HandlerContext, resource: AgentConfig) -> None:
            agent_config = ctx.get("map")
            del agent_config[resource.agentname]
            self._set_map(agent_config)

        def update_resource(
            self, ctx: HandlerContext, changes: dict, resource: AgentConfig
        ) -> None:
            agent_config = ctx.get("map")
            agent_config[resource.agentname] = resource.uri
            self._set_map(agent_config)

    waiter = Condition()

    async def wait_for_done_with_waiters(
        client,
        env_id,
        version,
        wait_for_this_amount_of_resources_in_done=None,
        timeout=10,
    ):
        # unhang waiters
        result = await client.get_version(env_id, version)
        assert result.code == 200
        now = time.time()
        logger.info(
            "waiting with waiters, %s resources done",
            result.result["model"]["done"],
        )
        while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
            if now + timeout < time.time():
                raise Exception("Timeout")
            if (
                wait_for_this_amount_of_resources_in_done
                and result.result["model"]["done"]
                - wait_for_this_amount_of_resources_in_done
                >= 0
            ):
                break
            result = await client.get_version(env_id, version)
            logger.info(
                "waiting with waiters, %s resources done",
                result.result["model"]["done"],
            )
            waiter.acquire()
            waiter.notifyAll()
            waiter.release()
            await asyncio.sleep(0.1)

        return result

    async def wait_for_condition_with_waiters(wait_condition, timeout=10):
        """
            Wait until wait_condition() returns false
        """
        now = time.time()
        while wait_condition():
            if now + timeout < time.time():
                raise Exception("Timeout")
            logger.info("waiting with waiters")
            waiter.acquire()
            waiter.notifyAll()
            waiter.release()
            await asyncio.sleep(0.1)

    @provider("test::Wait", name="test_wait")
    class Wait(ResourceHandler):
        def __init__(self, agent, io=None):
            super().__init__(agent, io)
            self.traceid = uuid.uuid4()

        def check_resource(self, ctx, resource):
            current = resource.clone()
            current.purged = not Provider.isset(
                resource.id.get_agent_name(), resource.key
            )

            if not current.purged:
                current.value = Provider.get(resource.id.get_agent_name(), resource.key)
            else:
                current.value = None

            return current

        def do_changes(self, ctx, resource, changes):
            logger.info("Hanging waiter %s", self.traceid)
            waiter.acquire()
            notified_before_timeout = waiter.wait(timeout=10)
            waiter.release()
            if not notified_before_timeout:
                raise Exception("Timeout occured")
            logger.info("Releasing waiter %s", self.traceid)
            if "purged" in changes:
                if changes["purged"]["desired"]:
                    Provider.delete(resource.id.get_agent_name(), resource.key)
                    ctx.set_purged()
                else:
                    Provider.set(
                        resource.id.get_agent_name(), resource.key, resource.value
                    )
                    ctx.set_created()

            if "value" in changes:
                Provider.set(resource.id.get_agent_name(), resource.key, resource.value)
                ctx.set_updated()

    @provider("test::WaitEvent", name="test_wait_Event")
    class WaitE(Provider):
        def __init__(self, agent, io=None):
            super().__init__(agent, io)
            self.traceid = uuid.uuid4()

        def check_resource(self, ctx, resource):
            current = resource.clone()
            current.purged = not Provider.isset(
                resource.id.get_agent_name(), resource.key
            )

            if not current.purged:
                current.value = Provider.get(resource.id.get_agent_name(), resource.key)
            else:
                current.value = None

            return current

        def process_events(self, ctx, resource, events):
            logger.info("Hanging Event waiter %s", self.traceid)
            waiter.acquire()
            notified_before_timeout = waiter.wait(timeout=10)
            waiter.release()
            if not notified_before_timeout:
                raise Exception("Timeout")
            logger.info("Releasing Event waiter %s", self.traceid)

    yield ResourceContainer(
        Provider=Provider,
        wait_for_done_with_waiters=wait_for_done_with_waiters,
        waiter=waiter,
        wait_for_condition_with_waiters=wait_for_condition_with_waiters,
    )
    Provider.reset()



import logging

import pytest

from agent_server.conftest import (
    get_resource,
    _deploy_resources,
    get_agent,
)
from utils import retry_limited, AsyncClosing, log_contains, log_doesnt_contain


@pytest.mark.asyncio(timeout=150)
async def test_deploy_trigger(
    server, client, resource_container, environment, caplog, no_agent_backoff
):
    """
       Test deployment of empty model
    """
    caplog.set_level(logging.INFO)
    async with AsyncClosing(get_agent(server, environment, "agent1", "agent5")):

        version = 5

        resources = [
            get_resource(version, agent="agent1"),
            get_resource(version, agent="agent2"),
            get_resource(version, agent="agent3"),
        ]

        await _deploy_resources(client, environment, resources, version, False)

        async def verify(
            result,
            a1=0,
            code=200,
            warnings=["Could not reach agents named [agent2,agent3]"],
            agents=["agent1"],
        ):
            assert result.code == code

            def is_deployed():
                return resource_container.Provider.readcount("agent1", "key1") == a1

            await retry_limited(is_deployed, 1)
            log_contains(
                caplog,
                "agent",
                logging.INFO,
                f"Agent agent1 got a trigger to update in environment {environment}",
            )
            log_doesnt_contain(
                caplog,
                "agent",
                logging.INFO,
                f"Agent agent5 got a trigger to update in environment {environment}",
            )

            assert result.result["agents"] == agents
            if warnings:
                assert sorted(result.result["metadata"]["warnings"]) == sorted(warnings)
            caplog.clear()

        async def verify_failed(
            result,
            code=400,
            message="",
            warnings=["Could not reach agents named [agent2,agent3]"],
        ):
            assert result.code == code

            log_doesnt_contain(caplog, "agent", logging.INFO, "got a trigger to update")
            if warnings:
                assert sorted(result.result["metadata"]["warnings"]) == sorted(warnings)
            assert result.result["message"] == message
            caplog.clear()

        # normal
        result = await client.deploy(environment)
        await verify(result, a1=1)

        # only agent1
        result = await client.deploy(environment, agents=["agent1"])
        await verify(result, a1=2, warnings=None)

        # only agent5 (not in model)
        result = await client.deploy(environment, agents=["agent5"])
        await verify_failed(
            result,
            404,
            "No agent could be reached",
            warnings=["Model version 5 does not contain agents named [agent5]"],
        )

        # only agent2 (not alive)
        result = await client.deploy(environment, agents=["agent2"])
        await verify_failed(
            result,
            404,
            "No agent could be reached",
            warnings=["Could not reach agents named [agent2]"],
        )

        # All of it
        result = await client.deploy(environment, agents=["agent1", "agent2", "agent5"])
        await verify(
            result,
            a1=3,
            agents=["agent1"],
            warnings=[
                "Could not reach agents named [agent2]",
                "Model version 5 does not contain agents named [agent5]",
            ],
        )



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import logging

from inmanta.ast.statements import ReferenceStatement
from inmanta.execute.runtime import ResultVariable, Waiter
from inmanta.execute.proxy import UnsetException, UnknownException
from inmanta.execute.util import Unknown
from inmanta.ast import RuntimeException, WrappingRuntimeException, ExternalException
from inmanta import plugins


LOGGER = logging.getLogger(__name__)


class FunctionCall(ReferenceStatement):
    """
        This class models a call to a function

        @param name: The name of the function that needs to be called
        @param arguments: A list of arguments

        uses:          args
        provides:      return value
        contributes:
    """

    def __init__(self, name, arguments):
        ReferenceStatement.__init__(self, arguments)
        self.name = name
        self.arguments = arguments

    def normalize(self):
        ReferenceStatement.normalize(self)
        self.function = self.namespace.get_type(self.name)
        self.allow_unknown = self.function.opts["allow_unknown"]

    def requires_emit(self, resolver, queue):
        sub = ReferenceStatement.requires_emit(self, resolver, queue)
        # add lazy vars
        temp = ResultVariable()
        FunctionUnit(queue, resolver, temp, sub, self)
        return {self: temp}

    def execute(self, requires, resolver, queue):
        return requires[self]

    def execute_direct(self, requires):
        function = self.function
        arguments = [a.execute_direct(requires) for a in self.arguments]
        no_unknows = function.check_args(arguments)

        if not no_unknows and not self.allow_unknown:
            raise RuntimeException("Received unknown value during direct execution")

        if function._context is not -1:
            raise RuntimeException("Context Aware functions are not allowed in direct execution")

        if function.opts["emits_statements"]:
            raise RuntimeException("emits_statements functions are not allowed in direct execution")
        else:
            try:
                return function(*arguments)
            except Exception as e:
                raise WrappingRuntimeException(self, "Exception in direct execution for plugin %s" % self.name, e)

    def resume(self, requires, resolver, queue, result):
        """
            Evaluate this statement.
        """
        # get the object to call the function on
        function = self.function
        arguments = [a.execute(requires, resolver, queue) for a in self.arguments]
        no_unknows = function.check_args(arguments)

        if not no_unknows and not self.allow_unknown:
            result.set_value(Unknown(self), self.location)
            return

        if function._context is not -1:
            arguments.insert(function._context, plugins.Context(resolver, queue, self, result))

        if function.opts["emits_statements"]:
            function(*arguments)
        else:
            try:
                value = function(*arguments)
                result.set_value(value, self.location)
            except UnknownException as e:
                result.set_value(e.unknown, self.location)
            except UnsetException as e:
                raise e
            except RuntimeException as e:
                raise WrappingRuntimeException(self, "Exception in plugin %s" % self.name, e)
            except Exception as e:
                raise ExternalException(self, "Exception in plugin %s" % self.name, e)

    def __repr__(self):
        return "%s(%s)" % (self.name, ','.join([repr(a) for a in self.arguments]))

    def pretty_print(self):
        return "%s(%s)" % (self.name, ','.join([a.pretty_print() for a in self.arguments]))


class FunctionUnit(Waiter):

    def __init__(self, queue_scheduler, resolver, result: ResultVariable, requires, function: FunctionCall):
        Waiter.__init__(self, queue_scheduler)
        self.result = result
        result.set_provider(self)
        self.requires = requires
        self.function = function
        self.resolver = resolver
        self.queue_scheduler = queue_scheduler
        for r in requires.values():
            self.waitfor(r)
        self.ready(self)

    def execute(self):
        requires = {k: v.get_value() for (k, v) in self.requires.items()}
        try:
            self.function.resume(requires, self.resolver, self.queue_scheduler, self.result)
            self.done = True
        except UnsetException as e:
            LOGGER.debug("Unset value in python code in plugin %s." % self.function.function)
            self.waitfor(e.get_result_variable())
        except RuntimeException as e:
            e.set_statement(self.function)
            raise e

    def __repr__(self):
        return repr(self.function)



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from inmanta.ast.statements import DynamicStatement
from inmanta.ast.statements.assign import Assign
from inmanta.ast import TypeNotFoundException, RuntimeException, Namespace, Anchor
from typing import List
from inmanta.execute.runtime import Resolver, QueueScheduler


class BasicBlock(object):

    def __init__(self, namespace: Namespace, stmts: List[DynamicStatement]=[]) -> None:
        self.__stmts = []  # type: List[DynamicStatement]
        self.variables = []  # type: List[str]
        self.namespace = namespace

        for st in stmts:
            self.add(st)

    def get_stmts(self) -> List[DynamicStatement]:
        return self.__stmts

    def get_anchors(self) -> List[Anchor]:
        return [a for s in self.__stmts for a in s.get_anchors()]

    def add(self, stmt: DynamicStatement) -> None:
        self.__stmts.append(stmt)

    def get_variables(self) -> List[str]:
        return self.variables

    def add_var(self, name: str) -> None:
        self.variables.append(name)

    def normalize(self) -> None:
        assigns = [s for s in self.__stmts if isinstance(s, Assign)]  # type: List[Assign]
        self.variables = [s.name for s in assigns]

        for s in self.__stmts:
            try:
                s.normalize()
            except TypeNotFoundException as e:
                e.set_statement(s)
                raise e
        # not used yet
        # self.requires = set([require for s in self.__stmts for require in s.requires()])

        # self.external = self.requires - set(self.variables)

        # self.external_not_global = [x for x in self.external if "::" not in x]

#     def get_requires(self) -> List[str]:
#         return self.external

    def emit(self, resolver: Resolver, queue: QueueScheduler) -> None:
        for s in self.__stmts:
            try:
                s.emit(resolver, queue)
            except RuntimeException as e:
                e.set_statement(s)
                raise e



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from inmanta.config import Option, is_int, is_bool, is_time, is_str_opt, is_map
from inmanta.config import state_dir, log_dir
import logging


LOGGER = logging.getLogger(__name__)

# flake8: noqa: H904

#############################
# Database
#############################

db_host = Option("database", "host", "localhost", "Hostname or IP of the postgresql server")
db_port = Option("database", "port", 5432, "The port of the postgresql server", is_int)
db_name = Option("database", "name", "inmanta", "The name of the database on the postgresql server")
db_username = Option("database", "username", "postgres", "The username to access the database in the PostgreSQL server")
db_password = Option("database", "password", None, "The password that belong to the database user")

#############################
# server_rest_transport
#############################
transport_port = Option("server_rest_transport", "port", 8888, "The port on which the server listens for connections")

#############################
# Influxdb
#############################
influxdb_host = Option("influxdb", "host", "", "Hostname or IP of the influxdb server to send reports to")
influxdb_port = Option("influxdb", "port", 8086, "The port of the influxdb server", is_int)
influxdb_name = Option("influxdb", "name", "inmanta", "The name of the database on the influxdb server")
influxdb_username = Option("influxdb", "username", None, "The username to access the database in the influxdb server")
influxdb_password = Option("influxdb", "password", None, "The password that belong to the influxdb user")

influxdb_interval = Option("influxdb", "interval", 30, "Interval with which to report to influxdb", is_int)
influxdb_tags = Option(
    "influxdb",
    "tags",
    "",
    "a dict of tags to attach to all influxdb records in the form tag=value,tag=value",
    is_map
)

#############################
# server
#############################
server_enable_auth = Option("server", "auth", False, "Enable authentication on the server API", is_bool)

server_ssl_key = Option("server", "ssl_key_file", None,
                        "Server private key to use for this server Leave blank to disable SSL", is_str_opt)

server_ssl_cert = Option("server", "ssl_cert_file", None,
                         "SSL certificate file for the server key. Leave blank to disable SSL", is_str_opt)

server_ssl_ca_cert = Option("server", "ssl_ca_cert_file", None,
                            "The CA cert file required to validate the server ssl cert. This setting is used by the server"
                            "to correctly configure the compiler and agents that the server starts itself. If not set and "
                            "SSL is enabled, the server cert should be verifiable with the CAs installed in the OS.",
                            is_str_opt)

server_fact_expire = Option("server", "fact-expire", 3600,
                            "After how many seconds will discovered facts/parameters expire", is_time)


def default_fact_renew():
    """ server.fact-expire/3 """
    return int(server_fact_expire.get() / 3)


def validate_fact_renew(value):
    """ time; < server.fact-expire """
    out = int(value)
    if not out < server_fact_expire.get():
        LOGGER.warn("can not set fact_renew to %d, must be smaller than fact-expire (%d), using %d instead" %
                    (out, server_fact_expire.get(), default_fact_renew()))
        out = default_fact_renew()
    return out


server_fact_renew = Option("server", "fact-renew", default_fact_renew,
                           """After how many seconds will discovered facts/parameters be renewed?
                              This value needs to be lower than fact-expire""", validate_fact_renew)

server_fact_resource_block = Option("server", "fact-resource-block", 60,
                                    "Minimal time between subsequent requests for the same fact", is_time)

server_autrecompile_wait = Option("server", "auto-recompile-wait", 10,
                                  """The number of seconds to wait before the server may attempt to do a new recompile.
                                     Recompiles are triggered after facts updates for example.""", is_time)

server_purge_version_interval = Option("server", "purge-versions-interval", 3600,
                                       """The number of seconds between version purging,
                                          see :inmanta.config:option:`server.available-versions-to-keep`""", is_time)

server_version_to_keep = Option("server", "available-versions-to-keep", 10,
                                """On boot and at regular intervals the server will purge older versions.
                                   This is the number of most recent versions to keep available.""", is_int)

server_address = Option("server", "server_address", "localhost",
                        """The public ip address of the server.
                           This is required for example to inject the inmanta agent in virtual machines at boot time.""")

server_wait_after_param = Option("server", "wait-after-param", 5,
                                 "Time to wait before recompile after new paramters have been received", is_time)

agent_timeout = Option("server", "agent-timeout", 30,
                       "Time before an agent is considered to be offline", is_time)

server_delete_currupt_files = Option("server", "delete_currupt_files", True,
                                     "The server logs an error when it detects a file got corrupted. When set to true, the "
                                     "server will also delete the file, so on subsequent compiles the missing file will be "
                                     "recreated.", is_bool)

server_purge_resource_action_logs_interval = Option("server", "purge-resource-action-logs-interval", 3600,
                                                    "The number of seconds between resource-action log purging", is_time)

server_resource_action_log_prefix = Option(
    "server",
    "resource_action_log_prefix",
    "resource-actions-",
    "File prefix in log-dir, containing the resource-action logs. The after the prefix the environment uuid and .log is added",
    is_str_opt
)


#############################
# Dashboard
#############################

dash_enable = Option("dashboard", "enabled", True, "Determines whether the server should host the dashboard or not", is_bool)

dash_path = Option("dashboard", "path", "/usr/share/inmanta/dashboard",
                   "The path on the local file system where the dashboard can be found")

dash_realm = Option("dashboard", "realm", "inmanta", "The realm to use for keycloak authentication.")
dash_auth_url = Option("dashboard", "auth_url", None, "The auth url of the keycloak server to use.")
dash_client_id = Option("dashboard", "client_id", None, "The client id configured in keycloak for this application.")
# LCM support should move to a server extension
dash_lcm_enable = Option("dashboard", "lcm", False, "Enable lifecycle manager in the dashboard", is_bool)


def default_hangtime():
    """ server.agent-timeout*3/4 """
    return str(int(agent_timeout.get() * 3 / 4))


agent_hangtime = Option("server", "agent-hold", default_hangtime,
                        "Maximal time the server will hold an agent heartbeat call", is_time)



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging
import time
import asyncio
import inspect


async def retry_limited(fun, timeout):
    async def fun_wrapper():
        if inspect.iscoroutinefunction(fun):
            return (await fun())
        else:
            return fun()

    start = time.time()
    while time.time() - start < timeout and not (await fun_wrapper()):
        await asyncio.sleep(0.1)
    if not (await fun_wrapper()):
        raise AssertionError("Bounded wait failed")


UNKWN = object()


def assert_equal_ish(minimal, actual, sortby=[]):
    if isinstance(minimal, dict):
        for k in minimal.keys():
            assert_equal_ish(minimal[k], actual[k], sortby)
    elif isinstance(minimal, list):
        assert len(minimal) == len(actual), "list not equal %s != %s" % (minimal, actual)
        if len(sortby) > 0:
            def keyfunc(val):
                if not isinstance(val, dict):
                    return val
                key = [str(val[x]) for x in sortby if x in val]
                return '_'.join(key)
            actual = sorted(actual, key=keyfunc)
        for (m, a) in zip(minimal, actual):
            assert_equal_ish(m, a, sortby)
    elif minimal is UNKWN:
        return
    else:
        assert minimal == actual


def assert_graph(graph, expected):
    lines = ["%s: %s" % (f.id.get_attribute_value(), t.id.get_attribute_value()) for f in graph.values() for t in f.requires]
    lines = sorted(lines)

    elines = [x.strip() for x in expected.split("\n")]
    elines = sorted(elines)

    assert elines == lines, (lines, elines)


class AsyncClosing(object):

    def __init__(self, awaitable):
        self.awaitable = awaitable

    async def __aenter__(self):
        self.closable = await self.awaitable
        return object

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.closable.stop()


def no_error_in_logs(caplog, levels=[logging.ERROR]):
    for logger_name, log_level, message in caplog.record_tuples:
        assert log_level not in levels, message


def log_contains(caplog, loggerpart, level, msg):
    close = []
    for logger_name, log_level, message in caplog.record_tuples:
        if msg in message:
            if loggerpart in logger_name and level == log_level:
                return
            else:
                close.append((logger_name, log_level, message))
    if close:
        print("found nearly matching log entry")
        for logger_name, log_level, message in close:
            print(logger_name, log_level, message)
        print("------------")

    assert False


def log_doesnt_contain(caplog, loggerpart, level, msg):
    for logger_name, log_level, message in caplog.record_tuples:
        if loggerpart in logger_name and level == log_level and msg in message:
            assert False


def log_index(caplog, loggerpart, level, msg, after=0):
    """Find a log in line in the captured log, return the index of the first occurrence

       :param after: only consider records after the given index"""
    close = []
    for i, (logger_name, log_level, message) in enumerate(caplog.record_tuples[after:]):
        if msg in message:
            if loggerpart in logger_name and level == log_level:
                return i + after
            else:
                close.append((logger_name, log_level, message))

    if close:
        print("found nearly matching")
        for logger_name, log_level, message in close:
            print(logger_name, log_level, message)
        print("------------")

    assert False


class LogSequence(object):

    def __init__(self, caplog, index=0):
        self.caplog = caplog
        self.index = index

    def log_contains(self, loggerpart, level, msg):
        index = log_index(self.caplog, loggerpart, level, msg, self.index)
        return LogSequence(self.caplog, index)



import logging
import time

from agent_server.conftest import get_resource, get_agent
from inmanta import config
import pytest

from utils import retry_limited, log_index


@pytest.mark.asyncio
async def test_agent_disconnect(
    resource_container, environment, server, client, async_finalizer, caplog
):
    caplog.set_level(logging.INFO)
    config.Config.set("config", "server-timeout", "1")
    config.Config.set("config", "agent-reconnect-delay", "1")
    config.Config.set("config", "agent-deploy-interval", "0")
    config.Config.set("config", "agent-repair-interval", "0")

    version = int(time.time())
    result = await client.put_version(
        tid=environment,
        version=version,
        resources=[get_resource(version)],
        unknowns=[],
        version_info={},
    )
    assert result.code == 200

    result = await client.release_version(environment, version, False)
    assert result.code == 200

    agent = await get_agent(server, environment, "agent1")
    async_finalizer.add(agent.stop)

    await server.stop()

    def disconnected():
        return not agent._instances["agent1"]._enabled

    await retry_limited(disconnected, 1)

    i = log_index(
        caplog,
        "inmanta.agent.agent.agent1",
        logging.INFO,
        "Agent assuming primary role for agent1",
    )
    i = log_index(
        caplog,
        "inmanta.agent.agent",
        logging.WARNING,
        "Connection to server lost, taking agents offline",
        i,
    )
    log_index(
        caplog,
        "inmanta.agent.agent.agent1",
        logging.INFO,
        "Agent agent1 stopped because Connection to server lost",
        i,
    )



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import sys
import os
import site
import subprocess
import tempfile
import hashlib
import logging
import re

import pkg_resources
from subprocess import CalledProcessError
import venv


LOGGER = logging.getLogger(__name__)


class VirtualEnv(object):
    """
        Creates and uses a virtual environment for this process
    """
    _egg_fragment_re = re.compile(r"#egg=(?P<name>[^&]*)")
    _at_fragment_re = re.compile(r"^(?P<name>[^@]+)@(?P<req>.+)")

    def __init__(self, env_path):
        LOGGER.info("Creating new virtual environment in %s", env_path)
        self.env_path = env_path
        self.virtual_python = None
        self.__cache_done = set()

        self._old = {}

    def init_env(self):
        """
            Init the virtual environment
        """
        python_name = os.path.basename(sys.executable)

        # check if the virtual env exists
        if sys.platform == "win32":
            python_bin = os.path.join(self.env_path, "Scripts", python_name)
        else:
            python_bin = os.path.join(self.env_path, "bin", python_name)

        if not os.path.exists(python_bin):
            # venv requires some care when the .env folder already exists
            # https://docs.python.org/3/library/venv.html
            if not os.path.exists(self.env_path):
                path = self.env_path
            else:
                # venv has problems with symlinks
                path = os.path.realpath(self.env_path)

            # --clear is required in python prior to 3.4 if the folder already exists
            try:
                venv.create(path, clear=True, with_pip=True)
            except CalledProcessError as e:
                LOGGER.exception("Unable to create new virtualenv at %s (%s)", self.env_path, e.stdout.decode())
                return False
            except Exception:
                LOGGER.exception("Unable to create new virtualenv at %s", self.env_path)
                return False
            LOGGER.debug("Created a new virtualenv at %s", self.env_path)

        # set the path to the python and the pip executables
        self.virtual_python = python_bin
        return True

    def use_virtual_env(self):
        """
            Use the virtual environment
        """
        if not self.init_env():
            raise Exception("Unable to init virtual environment")

        self.activate_that()

        # patch up pkg
        pkg_resources.working_set = pkg_resources.WorkingSet._build_master()

    def activate_that(self):
        # adapted from https://github.com/pypa/virtualenv/blob/master/virtualenv_embedded/activate_this.py
        # MIT license
        # Copyright (c) 2007 Ian Bicking and Contributors
        # Copyright (c) 2009 Ian Bicking, The Open Planning Project
        # Copyright (c) 2011-2016 The virtualenv developers

        if sys.platform == "win32":
            binpath = os.path.abspath(os.path.join(self.env_path, "Scripts"))
            base = os.path.dirname(binpath)
            site_packages = os.path.join(base, "Lib", "site-packages")
        else:
            binpath = os.path.abspath(os.path.join(self.env_path, "bin"))
            base = os.path.dirname(binpath)
            site_packages = os.path.join(base, "lib", "python%s" % sys.version[:3], "site-packages")

        old_os_path = os.environ.get("PATH", "")
        os.environ["PATH"] = binpath + os.pathsep + old_os_path
        prev_sys_path = list(sys.path)

        site.addsitedir(site_packages)
        sys.real_prefix = sys.prefix
        sys.prefix = base
        # Move the added items to the front of the path:
        new_sys_path = []
        for item in list(sys.path):
            if item not in prev_sys_path:
                new_sys_path.append(item)
                sys.path.remove(item)
        sys.path[:0] = new_sys_path

    def _parse_line(self, req_line: str) -> tuple:
        """
            Parse the requirement line
        """
        at = VirtualEnv._at_fragment_re.search(req_line)
        if at is not None:
            d = at.groupdict()
            return d["name"], d["req"] + "#egg=" + d["name"]

        egg = VirtualEnv._egg_fragment_re.search(req_line)
        if egg is not None:
            d = egg.groupdict()
            return d["name"], req_line

        return None, req_line

    def _gen_requirements_file(self, requirements_list) -> str:
        modules = {}
        for req in requirements_list:
            name, req_spec = self._parse_line(req)

            if name is None:
                name = req

            url = None
            version = None
            try:
                # this will fail is an url is supplied
                parsed_req = list(pkg_resources.parse_requirements(req_spec))
                if len(parsed_req) > 0:
                    item = parsed_req[0]
                    if hasattr(item, "name"):
                        name = item.name
                    elif hasattr(item, "unsafe_name"):
                        name = item.unsafe_name
                    version = item.specs
                    if hasattr(item, "url"):
                        url = item.url
            except pkg_resources.RequirementParseError:
                url = req_spec

            if name not in modules:
                modules[name] = {"name": name, "version": []}

            if version is not None:
                modules[name]["version"].extend(version)

            if url is not None:
                modules[name]["url"] = url

        requirements_file = ""
        for module, info in modules.items():
            version_spec = ""
            if len(info["version"]) > 0:
                version_spec = " " + (", ".join(["%s %s" % (a, b) for a, b in info["version"]]))

            if "url" in info:
                module = info["url"]

            requirements_file += module + version_spec + "\n"

        return requirements_file

    def _install(self, requirements_list: []) -> None:
        """
            Install requirements in the given requirements file
        """
        requirements_file = self._gen_requirements_file(requirements_list)

        try:
            fdnum, path = tempfile.mkstemp()
            fd = os.fdopen(fdnum, "w+")
            fd.write(requirements_file)
            fd.close()

            cmd = [self.virtual_python, "-m", "pip", "install", "-r", path]
            output = b""
            try:
                output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            except CalledProcessError as e:
                LOGGER.debug("%s: %s", cmd, e.output.decode())
                LOGGER.debug("requirements: %s", requirements_file)
                raise
            except Exception:
                LOGGER.debug("%s: %s", cmd, output.decode())
                LOGGER.debug("requirements: %s", requirements_file)
                raise
            else:
                LOGGER.debug("%s: %s", cmd, output.decode())

        finally:
            if os.path.exists(path):
                os.remove(path)

        pkg_resources.working_set = pkg_resources.WorkingSet._build_master()

    def _read_current_requirements_hash(self):
        """
            Return the hash of the requirements file used to install the current environment
        """
        path = os.path.join(self.env_path, "requirements.sha1sum")
        if not os.path.exists(path):
            return ""

        with open(path, "r") as fd:
            return fd.read().strip()

    def _set_current_requirements_hash(self, new_hash):
        """
            Set the current requirements hahs
        """
        path = os.path.join(self.env_path, "requirements.sha1sum")
        with open(path, "w+") as fd:
            fd.write(new_hash)

    def install_from_list(self, requirements_list: list, detailed_cache=False, cache=True) -> None:
        """
            Install requirements from a list of requirement strings
        """
        requirements_list = sorted(requirements_list)

        if detailed_cache:
            requirements_list = sorted(list(set(requirements_list) - self.__cache_done))
            if len(requirements_list) == 0:
                return

        # hash it
        sha1sum = hashlib.sha1()
        sha1sum.update("\n".join(requirements_list).encode())
        new_req_hash = sha1sum.hexdigest()

        current_hash = self._read_current_requirements_hash()

        if new_req_hash == current_hash and cache:
            return

        self._install(requirements_list)
        self._set_current_requirements_hash(new_req_hash)
        for x in requirements_list:
            self.__cache_done.add(x)



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import inspect
import enum
import uuid
import logging
import json
import gzip
import io
import time

import jwt

from tornado import web
from urllib import parse
from typing import Any, Dict, List, Optional, Union, Tuple, Set, Callable, Generator, cast, TYPE_CHECKING  # noqa: F401

from inmanta import execute, const, util
from inmanta import config as inmanta_config
from inmanta.types import JsonType
from . import exceptions

if TYPE_CHECKING:
    from .endpoints import CallTarget


LOGGER: logging.Logger = logging.getLogger(__name__)


class ArgOption(object):
    """
        Argument options to transform arguments before dispatch
    """

    def __init__(self, header: Optional[str] = None, reply_header: bool = True, getter: Optional[Generator] = None) -> None:
        """
            :param header: Map this argument to a header with the following name.
            :param reply_header: If the argument is mapped to a header, this header will also be included in the reply
            :param getter: Call this method after validation and pass its return value to the method call. This may change the
                           type of the argument. This method can raise an HTTPException to return a 404 for example.
        """
        self.header = header
        self.reply_header = reply_header
        self.getter = getter


class Request(object):
    """
        A protocol request
    """

    def __init__(self, url: str, method: str, headers: Dict[str, str], body: Dict[str, Any]) -> None:
        self._url = url
        self._method = method
        self._headers = headers
        self._body = body
        self._reply_id: Optional[uuid.UUID] = None

    @property
    def body(self) -> Dict[str, Any]:
        return self._body

    @property
    def url(self) -> str:
        return self._url

    @property
    def headers(self) -> Dict[str, str]:
        return self._headers

    @property
    def method(self) -> str:
        return self._method

    def set_reply_id(self, reply_id: uuid.UUID) -> None:
        self._reply_id = reply_id

    def get_reply_id(self) -> Optional[uuid.UUID]:
        return self._reply_id

    reply_id = property(get_reply_id, set_reply_id)

    def to_dict(self) -> Dict[str, Any]:
        return_dict: Dict[str, Any] = {"url": self._url, "headers": self._headers, "body": self._body, "method": self._method}
        if self._reply_id is not None:
            return_dict["reply_id"] = self._reply_id

        return return_dict

    @classmethod
    def from_dict(cls, value: Dict[str, Any]) -> "Request":
        reply_id: Optional[str] = None
        if "reply_id" in value:
            reply_id = cast(str, value["reply_id"])
            del value["reply_id"]

        req = Request(**value)

        if reply_id is not None:
            req.reply_id = uuid.UUID(reply_id)

        return req


class Response(object):
    """
        A response object of a call
    """

    def __init__(self, status_code: int, headers: Dict[str, str], body: Optional[Dict[str, Any]] = None) -> None:
        self._status_code = status_code
        self._headers = headers
        self._body = body

    @property
    def body(self) -> Optional[Dict[str, Any]]:
        return self._body

    @property
    def headers(self) -> Dict[str, str]:
        return self._headers

    @property
    def status_code(self) -> int:
        return self._status_code


class MethodProperties(object):
    """
        This class stores the information from a method definition
    """

    methods: Dict[str, "MethodProperties"] = {}

    def __init__(
        self,
        function: Callable[..., Dict[str, Any]],
        method_name: str,
        index: bool,
        id: bool,
        operation: str,
        reply: bool,
        arg_options: Dict[str, ArgOption],
        timeout: Optional[int],
        server_agent: bool,
        api: Optional[bool],
        agent_server: bool,
        validate_sid: Optional[bool],
        client_types: List[str],
        api_version: int,
    ) -> None:
        """
            Decorator to identify a method as a RPC call. The arguments of the decorator are used by each transport to build
            and model the protocol.

            :param method_name: The method name in the url
            :param index: A method that returns a list of resources. The url of this method is only the method/resource name.
            :param id: This method requires an id of a resource. The python function should have an id parameter.
            :param operation: The type of HTTP operation (verb)
            :param timeout: nr of seconds before request it terminated
            :param api This is a call from the client to the Server (True if not server_agent and not agent_server)
            :param server_agent: This is a call from the Server to the Agent (reverse http channel through long poll)
            :param agent_server: This is a call from the Agent to the Server
            :param validate_sid: This call requires a valid session, true by default if agent_server and not api
            :param client_types: The allowed client types for this call
            :param arg_options: Options related to arguments passed to the method. The key of this dict is the name of the arg
                                to which the options apply.
            :param api_version: The version of the api this method belongs to

        """
        if api is None:
            api = not server_agent and not agent_server

        if validate_sid is None:
            validate_sid = agent_server and not api

        self._method_name = method_name
        self._index = index
        self._id = id
        self._operation = operation
        self._reply = reply
        self._arg_options = arg_options
        self._timeout = timeout
        self._server_agent = server_agent
        self._api: bool = api
        self._agent_server = agent_server
        self._validate_sid: bool = validate_sid
        self._client_types = client_types
        self._api_version = api_version
        self.function = function

        MethodProperties.methods[function.__name__] = self
        function.__method_properties__ = self

        # validate client types
        for ct in self._client_types:
            if ct not in const.VALID_CLIENT_TYPES:
                raise Exception("Invalid client type %s specified for function %s" % (ct, function))

    @property
    def operation(self) -> str:
        return self._operation

    @property
    def arg_options(self) -> Dict[str, ArgOption]:
        return self._arg_options

    @property
    def timeout(self) -> Optional[int]:
        return self._timeout

    @property
    def id(self) -> bool:
        return self._id

    @property
    def validate_sid(self) -> bool:
        return self._validate_sid

    @property
    def agent_server(self) -> bool:
        return self._agent_server

    @property
    def reply(self) -> bool:
        return self._reply

    @property
    def client_types(self) -> List[str]:
        return self._client_types

    def get_call_headers(self) -> Set[str]:
        """
            Returns the set of headers required to create call
        """
        headers = set()
        headers.add("Authorization")

        for arg in self._arg_options.values():
            if arg.header is not None:
                headers.add(arg.header)

        return headers

    def get_listen_url(self) -> str:
        """
            Create a listen url for this method
        """
        url = "/api/v%d" % self._api_version

        if self._id:
            url += "/%s/(?P<id>[^/]+)" % self._method_name
        elif self._index:
            url += "/%s" % self._method_name
        else:
            url += "/%s" % self._method_name

        return url

    def get_call_url(self, msg: Dict[str, str]) -> str:
        """
             Create a calling url for the client
        """
        url = "/api/v%d" % self._api_version

        if self._id:
            url += "/%s/%s" % (self._method_name, parse.quote(str(msg["id"]), safe=""))
        elif self._index:
            url += "/%s" % self._method_name
        else:
            url += "/%s" % self._method_name

        return url

    def build_call(self, args: List, kwargs: Dict[str, Any] = {}) -> Request:
        """
            Build a call from the given arguments. This method returns the url, headers, and body for the call.
        """
        # create the message
        msg: Dict[str, Any] = dict(kwargs)

        # map the argument in arg to names
        argspec = inspect.getfullargspec(self.function)
        for i in range(len(args)):
            msg[argspec.args[i]] = args[i]

        url = self.get_call_url(msg)

        headers = {}

        for arg_name in list(msg.keys()):
            if isinstance(msg[arg_name], enum.Enum):  # Handle enum values "special"
                msg[arg_name] = msg[arg_name].name

            if arg_name in self.arg_options:
                opts = self.arg_options[arg_name]
                if opts.header:
                    headers[opts.header] = str(msg[arg_name])
                    del msg[arg_name]

        if self.operation not in ("POST", "PUT", "PATCH"):
            qs_map = msg.copy()
            if "id" in qs_map:
                del qs_map["id"]

            # encode arguments in url
            if len(qs_map) > 0:
                url += "?" + parse.urlencode(qs_map)

            body = None
        else:
            body = msg

        return Request(url=url, method=self.operation, headers=headers, body=body)


class UrlMethod(object):
    """
        This class holds the method definition together with the API (url, method) information

        :param properties: The properties of this method
        :param endpoint: The object on which this method is defined
        :param handler: The method to call on the endpoint
        :param method_name: The name of the method to call on the endpoint
    """

    def __init__(
        self,
        properties: MethodProperties,
        slice: "CallTarget",
        handler: Callable[..., Dict[int, Dict[str, Any]]],
        method_name: str,
    ):
        self._properties = properties
        self._handler = handler
        self._slice = slice
        self._method_name = method_name

    @property
    def properties(self) -> MethodProperties:
        return self._properties

    @property
    def handler(self) -> Callable[..., Dict[int, Dict[str, Any]]]:
        return self._handler

    @property
    def endpoint(self) -> "CallTarget":
        return self._slice

    @property
    def method_name(self) -> str:
        return self._method_name


# Util functions
def custom_json_encoder(o: object) -> Union[Dict, str, List]:
    """
        A custom json encoder that knows how to encode other types commonly used by Inmanta
    """
    if isinstance(o, execute.util.Unknown):
        return const.UNKNOWN_STRING

    # handle common python types
    return util.custom_json_encoder(o)


def attach_warnings(code: int, value: JsonType, warnings: Optional[List[str]]) -> Tuple[int, JsonType]:
    if warnings:
        meta = value.setdefault("metadata", {})
        warns = meta.setdefault("warnings", [])
        warns.extend(warnings)
    return code, value


def json_encode(value: JsonType) -> str:
    # see json_encode in tornado.escape
    return json.dumps(value, default=custom_json_encoder).replace("</", "<\\/")


def gzipped_json(value: JsonType) -> Tuple[bool, Union[bytes, str]]:
    json_string = json_encode(value)
    if len(json_string) < web.GZipContentEncoding.MIN_LENGTH:
        return False, json_string

    gzip_value = io.BytesIO()
    gzip_file = gzip.GzipFile(mode="w", fileobj=gzip_value, compresslevel=web.GZipContentEncoding.GZIP_LEVEL)

    gzip_file.write(json_string.encode())
    gzip_file.close()

    return True, gzip_value.getvalue()


def shorten(msg: str, max_len: int = 10) -> str:
    if len(msg) < max_len:
        return msg
    return msg[0: max_len - 3] + "..."


def encode_token(client_types: List[str], environment: str = None, idempotent: bool = False, expire: float = None) -> str:
    cfg = inmanta_config.AuthJWTConfig.get_sign_config()

    payload = {"iss": cfg.issuer, "aud": [cfg.audience], const.INMANTA_URN + "ct": ",".join(client_types)}

    if not idempotent:
        payload["iat"] = int(time.time())

        if cfg.expire > 0:
            payload["exp"] = int(time.time() + cfg.expire)
        elif expire is not None:
            payload["exp"] = int(time.time() + expire)

    if environment is not None:
        payload[const.INMANTA_URN + "env"] = environment

    return jwt.encode(payload, cfg.key, cfg.algo).decode()


def decode_token(token: str) -> Dict[str, str]:
    try:
        # First decode the token without verification
        header = jwt.get_unverified_header(token)
        payload = jwt.decode(token, verify=False)
    except Exception:
        raise exceptions.AccessDeniedException("Unable to decode provided JWT bearer token.")

    if "iss" not in payload:
        raise exceptions.AccessDeniedException("Issuer is required in token to validate.")

    cfg = inmanta_config.AuthJWTConfig.get_issuer(payload["iss"])
    if cfg is None:
        raise exceptions.AccessDeniedException("Unknown issuer for token")

    alg = header["alg"].lower()
    if alg == "hs256":
        key = cfg.key
    elif alg == "rs256":
        if "kid" not in header:
            raise exceptions.AccessDeniedException("A kid is required for RS256")
        kid = header["kid"]
        if kid not in cfg.keys:
            raise exceptions.AccessDeniedException(
                "The kid provided in the token does not match a known key. Check the jwks_uri or try "
                "restarting the server to load any new keys."
            )

        key = cfg.keys[kid]
    else:
        raise exceptions.AccessDeniedException("Algorithm %s is not supported." % alg)

    try:
        payload = dict(jwt.decode(token, key, audience=cfg.audience, algorithms=[cfg.algo]))
        ct_key = const.INMANTA_URN + "ct"
        payload[ct_key] = [x.strip() for x in payload[ct_key].split(",")]
    except Exception as e:
        raise exceptions.AccessDeniedException(*e.args)

    return payload


class Result(object):
    """
        A result of a method call
    """

    def __init__(self, code: int = 0, result: Dict[str, Any] = None):
        self._result = result
        self.code = code
        self._callback = None

    def get_result(self) -> Dict[str, Any]:
        """
            Only when the result is marked as available the result can be returned
        """
        if self.available():
            return self._result
        raise Exception("The result is not yet available")

    def set_result(self, value):
        if not self.available():
            self._result = value
            if self._callback:
                self._callback(self)

    def available(self) -> bool:
        return self._result is not None or self.code > 0

    def wait(self, timeout: int = 60) -> None:
        """
            Wait for the result to become available
        """
        count: float = 0
        while count < timeout:
            time.sleep(0.1)
            count += 0.1

    result = property(get_result, set_result)

    def callback(self, fnc):
        """
            Set a callback function that is to be called when the result is ready.
        """
        self._callback = fnc


class SessionManagerInterface(object):
    """
        An interface for a sessionmanager
    """
    def validate_sid(self, sid: uuid.UUID) -> bool:
        """
        Check if the given sid is a valid session
        :param sid: The session id
        :return: True if the session is valid
        """
        raise NotImplementedError()



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import os
import tempfile
import random
import string
import shutil
from tempfile import mktemp
import socket
import logging


import pytest

import utils
from inmanta import data, config
import inmanta.compiler as compiler
from inmanta.module import Project
from inmanta import resources
import inmanta.agent
from inmanta.agent import handler
from inmanta.agent.agent import Agent
from inmanta.ast import CompilerException
from click import testing
import inmanta.main
import re
from inmanta.server.bootloader import InmantaBootloader
from inmanta.server import SLICE_AGENT_MANAGER
from inmanta.export import cfg_env, unknown_parameters
import traceback
from tornado import process, netutil
import asyncio
from tornado.platform.asyncio import AnyThreadEventLoopPolicy
import asyncpg
from asyncpg.exceptions import DuplicateDatabaseError
from inmanta.postgresproc import PostgresProc
import sys
import pkg_resources
from typing import Optional, Dict
from inmanta import protocol
import pyformance
from pyformance.registry import MetricsRegistry

from inmanta.util import get_free_tcp_port

asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())


@pytest.fixture(scope="session", autouse=True)
def postgres_db(postgresql_proc):
    yield postgresql_proc


@pytest.fixture(scope="function")
async def postgresql_client(postgres_db, database_name):
    client = await asyncpg.connect(host=postgres_db.host, port=postgres_db.port, user=postgres_db.user, database=database_name)
    try:
        yield client
    finally:
        await client.close()


@pytest.fixture(scope="function")
async def init_dataclasses_and_load_schema(postgres_db, database_name, clean_reset):
    await data.connect(postgres_db.host, postgres_db.port, database_name, postgres_db.user, None)
    yield
    await data.disconnect()


@pytest.fixture(scope="function", autouse=True)
def deactive_venv():
    old_os_path = os.environ.get("PATH", "")
    old_prefix = sys.prefix
    old_path = sys.path

    yield

    os.environ["PATH"] = old_os_path
    sys.prefix = old_prefix
    sys.path = old_path
    pkg_resources.working_set = pkg_resources.WorkingSet._build_master()


def reset_metrics():
    pyformance.set_global_registry(MetricsRegistry())


@pytest.fixture(scope="function", autouse=True)
async def clean_reset(create_db, clean_db):
    reset_all_objects()
    config.Config._reset()
    yield
    config.Config._reset()
    reset_all_objects()


def reset_all_objects():
    resources.resource.reset()
    process.Subprocess.uninitialize()
    asyncio.set_child_watcher(None)
    reset_metrics()
    # No dynamic loading of commands at the moment, so no need to reset/reload
    # command.Commander.reset()
    handler.Commander.reset()
    Project._project = None
    unknown_parameters.clear()


@pytest.fixture(scope="function")
async def create_db(postgres_db, database_name):
    connection = await asyncpg.connect(host=postgres_db.host, port=postgres_db.port, user=postgres_db.user)
    try:
        await connection.execute("CREATE DATABASE " + database_name)
    except DuplicateDatabaseError:
        pass
    finally:
        await connection.close()


@pytest.fixture(scope="function")
async def clean_db(postgresql_client, create_db):
    """
        1) Truncated tables: All tables which are part of the inmanta schema, except for the schemaversion table. The version
                             number stored in the schemaversion table is read by the Inmanta server during startup.
        2) Dropped tables: All tables which are not part of the inmanta schema. Some tests create additional tables, which are
                           not part of the Inmanta schema. These should be cleaned-up before running a new test.
    """
    tables_in_db = await postgresql_client.fetch("SELECT table_name FROM information_schema.tables WHERE table_schema='public'")
    tables_in_db = [x["table_name"] for x in tables_in_db]
    tables_to_preserve = [x.table_name() for x in data._classes]
    tables_to_truncate = [x for x in tables_in_db if x != data.SchemaVersion.table_name() and x in tables_to_preserve]
    tables_to_drop = [x for x in tables_in_db if x not in tables_to_preserve]
    if tables_to_drop:
        drop_query = "DROP TABLE %s CASCADE" % ', '.join(tables_to_drop)
        await postgresql_client.execute(drop_query)
    if tables_to_truncate:
        truncate_query = "TRUNCATE %s CASCADE" % ', '.join(tables_to_truncate)
        await postgresql_client.execute(truncate_query)


@pytest.fixture(scope="function", autouse=True)
def restore_cwd():
    """
        Restore the current working directory after search test.
    """
    cwd = os.getcwd()
    yield
    os.chdir(cwd)


@pytest.fixture(scope="function")
def no_agent_backoff():
    backoff = inmanta.agent.agent.GET_RESOURCE_BACKOFF
    inmanta.agent.agent.GET_RESOURCE_BACKOFF = 0
    yield
    inmanta.agent.agent.GET_RESOURCE_BACKOFF = backoff


@pytest.fixture()
def free_socket():
    sock = netutil.bind_sockets(0, "127.0.0.1", family=socket.AF_INET)[0]
    yield sock
    sock.close()


@pytest.fixture()
def free_port():
    port = get_free_tcp_port()
    yield port


@pytest.fixture(scope="function", autouse=True)
def inmanta_config():
    config.Config.load_config()
    config.Config.set("auth_jwt_default", "algorithm", "HS256")
    config.Config.set("auth_jwt_default", "sign", "true")
    config.Config.set("auth_jwt_default", "client_types", "agent,compiler")
    config.Config.set("auth_jwt_default", "key", "rID3kG4OwGpajIsxnGDhat4UFcMkyFZQc1y3oKQTPRs")
    config.Config.set("auth_jwt_default", "expire", "0")
    config.Config.set("auth_jwt_default", "issuer", "https://localhost:8888/")
    config.Config.set("auth_jwt_default", "audience", "https://localhost:8888/")

    yield config.Config._get_instance()


@pytest.fixture(scope="session")
def database_name():
    ten_random_digits = ''.join(random.choice(string.digits) for _ in range(10))
    yield "inmanta" + ten_random_digits


@pytest.fixture(scope="function")
async def agent_multi(server_multi, environment_multi):
    agentmanager = server_multi.get_slice(SLICE_AGENT_MANAGER)

    config.Config.set("config", "agent-deploy-interval", "0")
    config.Config.set("config", "agent-repair-interval", "0")
    a = Agent(hostname="node1", environment=environment_multi, agent_map={"agent1": "localhost"}, code_loader=False)
    a.add_end_point_name("agent1")
    await a.start()
    await utils.retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    yield a

    await a.stop()


@pytest.fixture(scope="function")
async def agent(server, environment):
    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)

    config.Config.set("config", "agent-deploy-interval", "0")
    config.Config.set("config", "agent-repair-interval", "0")
    a = Agent(hostname="node1", environment=environment, agent_map={"agent1": "localhost"}, code_loader=False)
    a.add_end_point_name("agent1")
    await a.start()
    await utils.retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    yield a

    await a.stop()


@pytest.fixture(scope="function")
async def server(event_loop, inmanta_config, postgres_db, database_name, clean_reset):
    """
    :param event_loop: explicitly include event_loop to make sure event loop started before and closed after this fixture.
    May not be required
    """
    # fix for fact that pytest_tornado never set IOLoop._instance, the IOLoop of the main thread
    # causes handler failure

    reset_metrics()

    state_dir = tempfile.mkdtemp()

    port = get_free_tcp_port()
    config.Config.set("database", "name", database_name)
    config.Config.set("database", "host", "localhost")
    config.Config.set("database", "port", str(postgres_db.port))
    config.Config.set("config", "state-dir", state_dir)
    config.Config.set("config", "log-dir", os.path.join(state_dir, "logs"))
    config.Config.set("server_rest_transport", "port", port)
    config.Config.set("agent_rest_transport", "port", port)
    config.Config.set("compiler_rest_transport", "port", port)
    config.Config.set("client_rest_transport", "port", port)
    config.Config.set("cmdline_rest_transport", "port", port)
    config.Config.set("config", "executable", os.path.abspath(os.path.join(__file__, "../../src/inmanta/app.py")))
    config.Config.set("server", "agent-timeout", "10")
    config.Config.set("server", "auto-recompile-wait", "0")
    config.Config.set("agent", "agent-repair-interval", "0")

    ibl = InmantaBootloader()
    await ibl.start()

    yield ibl.restserver

    await asyncio.wait_for(ibl.stop(), 10)
    shutil.rmtree(state_dir)


@pytest.fixture(scope="function",
                params=[(True, True, False), (True, False, False), (False, True, False),
                        (False, False, False), (True, True, True)],
                ids=["SSL and Auth", "SSL", "Auth", "Normal", "SSL and Auth with not self signed certificate"])
async def server_multi(event_loop, inmanta_config, postgres_db, database_name, request, clean_reset):
    """
    :param event_loop: explicitly include event_loop to make sure event loop started before and closed after this fixture.
    May not be required
    """
    state_dir = tempfile.mkdtemp()

    ssl, auth, ca = request.param

    path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")

    if auth:
        config.Config.set("server", "auth", "true")

    for x, ct in [("server", None),
                  ("server_rest_transport", None),
                  ("agent_rest_transport", ["agent"]),
                  ("compiler_rest_transport", ["compiler"]),
                  ("client_rest_transport", ["api", "compiler"]),
                  ("cmdline_rest_transport", ["api"])]:
        if ssl and not ca:
            config.Config.set(x, "ssl_cert_file", os.path.join(path, "server.crt"))
            config.Config.set(x, "ssl_key_file", os.path.join(path, "server.open.key"))
            config.Config.set(x, "ssl_ca_cert_file", os.path.join(path, "server.crt"))
            config.Config.set(x, "ssl", "True")
        if ssl and ca:
            capath = os.path.join(path, "ca", "enduser-certs")

            config.Config.set(x, "ssl_cert_file", os.path.join(capath, "server.crt"))
            config.Config.set(x, "ssl_key_file", os.path.join(capath, "server.key.open"))
            config.Config.set(x, "ssl_ca_cert_file", os.path.join(capath, "server.chain"))
            config.Config.set(x, "ssl", "True")
        if auth and ct is not None:
            token = protocol.encode_token(ct)
            config.Config.set(x, "token", token)

    port = get_free_tcp_port()
    config.Config.set("database", "name", database_name)
    config.Config.set("database", "host", "localhost")
    config.Config.set("database", "port", str(postgres_db.port))
    config.Config.set("config", "state-dir", state_dir)
    config.Config.set("config", "log-dir", os.path.join(state_dir, "logs"))
    config.Config.set("server_rest_transport", "port", port)
    config.Config.set("agent_rest_transport", "port", port)
    config.Config.set("compiler_rest_transport", "port", port)
    config.Config.set("client_rest_transport", "port", port)
    config.Config.set("cmdline_rest_transport", "port", port)
    config.Config.set("config", "executable", os.path.abspath(os.path.join(__file__, "../../src/inmanta/app.py")))
    config.Config.set("server", "agent-timeout", "2")
    config.Config.set("agent", "agent-repair-interval", "0")
    config.Config.set("server", "auto-recompile-wait", "0")

    ibl = InmantaBootloader()
    await ibl.start()

    yield ibl.restserver

    await asyncio.wait_for(ibl.stop(), 10)

    shutil.rmtree(state_dir)


@pytest.fixture(scope="function")
def client(server):
    client = protocol.Client("client")
    yield client


@pytest.fixture(scope="function")
def client_multi(server_multi):
    client = protocol.Client("client")
    yield client


@pytest.fixture(scope="function")
def sync_client_multi(server_multi):
    client = protocol.SyncClient("client")
    yield client


@pytest.fixture(scope="function", autouse=True)
def capture_warnings():
    logging.captureWarnings(True)
    yield
    logging.captureWarnings(False)


@pytest.fixture(scope="function")
async def environment(client, server):
    """
        Create a project and environment. This fixture returns the uuid of the environment
    """
    result = await client.create_project("env-test")
    assert(result.code == 200)
    project_id = result.result["project"]["id"]

    result = await client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    cfg_env.set(env_id)

    yield env_id


@pytest.fixture(scope="function")
async def environment_multi(client_multi, server_multi):
    """
        Create a project and environment. This fixture returns the uuid of the environment
    """
    result = await client_multi.create_project("env-test")
    assert(result.code == 200)
    project_id = result.result["project"]["id"]

    result = await client_multi.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    cfg_env.set(env_id)

    yield env_id


@pytest.fixture(scope="session")
def write_db_update_file():

    def _write_db_update_file(schema_dir, schema_version, content_file):
        schema_updates_dir = os.path.join(schema_dir, data.DBSchema.DIR_NAME_INCREMENTAL_UPDATES)
        if not os.path.exists(schema_updates_dir):
            os.mkdir(schema_updates_dir)
        schema_update_file = os.path.join(schema_updates_dir, str(schema_version) + ".sql")
        with open(schema_update_file, 'w+') as f:
            f.write(content_file)
    yield _write_db_update_file


@pytest.fixture(scope="function")
def get_columns_in_db_table(postgresql_client):
    async def _get_columns_in_db_table(table_name):
        result = await postgresql_client.fetch("SELECT column_name "
                                               "FROM information_schema.columns "
                                               "WHERE table_schema='public' AND table_name='" + table_name + "'")
        return [r["column_name"] for r in result]
    return _get_columns_in_db_table


class KeepOnFail(object):

    def keep(self) -> "Optional[Dict[str, str]]":
        pass


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # we only look at actual failing test calls, not setup/teardown
    resources = {}
    if rep.when == "call" and rep.failed:
        for fixture in item.funcargs.values():
            if isinstance(fixture, KeepOnFail):
                msg = fixture.keep()
                if msg:
                    for label, res in msg.items():
                        resources[label] = res

        if resources:
            # we are behind report formatting, so write to report, not item
            rep.sections.append(("Resources Kept", "\n".join(
                ["%s %s" % (label, resource) for label, resource in resources.items()])))


async def off_main_thread(func):
    return await asyncio.get_event_loop().run_in_executor(None, func)


class SnippetCompilationTest(KeepOnFail):

    def setUpClass(self):
        self.libs = tempfile.mkdtemp()
        self.env = tempfile.mkdtemp()
        config.Config.load_config()
        self.cwd = os.getcwd()
        self.keep_shared = False

    def tearDownClass(self):
        if not self.keep_shared:
            shutil.rmtree(self.libs)
            shutil.rmtree(self.env)
        # reset cwd
        os.chdir(self.cwd)

    def setup_func(self):
        # init project
        self._keep = False
        self.project_dir = tempfile.mkdtemp()
        os.symlink(self.env, os.path.join(self.project_dir, ".env"))

    def tear_down_func(self):
        if not self._keep:
            shutil.rmtree(self.project_dir)

    def keep(self):
        self._keep = True
        self.keep_shared = True
        return {"env": self.env, "libs": self.libs, "project": self.project_dir}

    def setup_for_snippet(self, snippet, autostd=True):
        self.setup_for_snippet_external(snippet)

        Project.set(Project(self.project_dir, autostd=autostd))

    def setup_for_snippet_external(self, snippet):
        with open(os.path.join(self.project_dir, "project.yml"), "w") as cfg:
            cfg.write(
                """
            name: snippet test
            modulepath: [%s, %s]
            downloadpath: %s
            version: 1.0
            repo: ['https://github.com/inmanta/']"""
                % (self.libs,
                   os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "modules"),
                   self.libs))
        self.main = os.path.join(self.project_dir, "main.cf")
        with open(self.main, "w") as x:
            x.write(snippet)

    def do_export(self, include_status=False, do_raise=True):
        return self._do_export(deploy=False, include_status=include_status, do_raise=do_raise)

    def _do_export(self, deploy=False, include_status=False, do_raise=True):
        """
        helper function to allow actual export to be run an a different thread
        i.e. export.run must run off main thread to allow it to start a new ioloop for run_sync
        """
        templfile = mktemp("json", "dump", self.project_dir)

        class Options(object):
            pass

        options = Options()
        options.json = templfile if not deploy else None
        options.depgraph = False
        options.deploy = deploy
        options.ssl = False

        from inmanta.export import Exporter  # noqa: H307

        try:
            (types, scopes) = compiler.do_compile()
        except Exception:
            types, scopes = (None, None)
            if do_raise:
                raise
            else:
                traceback.print_exc()

        # Even if the compile failed we might have collected additional data such as unknowns. So
        # continue the export
        export = Exporter(options)

        return export.run(types, scopes, model_export=False, include_status=include_status)

    async def do_export_and_deploy(self, include_status=False, do_raise=True):
        return await off_main_thread(lambda: self._do_export(deploy=True, include_status=include_status, do_raise=do_raise))

    def setup_for_error(self, snippet, shouldbe):
        self.setup_for_snippet(snippet)
        try:
            compiler.do_compile()
            assert False, "Should get exception"
        except CompilerException as e:
            text = e.format_trace(indent="  ")
            print(text)
            shouldbe = shouldbe.format(dir=self.project_dir)
            assert shouldbe == text

    def setup_for_error_re(self, snippet, shouldbe):
        self.setup_for_snippet(snippet)
        try:
            compiler.do_compile()
            assert False, "Should get exception"
        except CompilerException as e:
            text = e.format_trace(indent="  ")
            print(text)
            shouldbe = shouldbe.format(dir=self.project_dir)
            assert re.search(shouldbe, text) is not None


@pytest.fixture(scope="session")
def snippetcompiler_global():
    ast = SnippetCompilationTest()
    ast.setUpClass()
    yield ast
    ast.tearDownClass()


@pytest.fixture(scope="function")
def snippetcompiler(snippetcompiler_global):
    snippetcompiler_global.setup_func()
    yield snippetcompiler_global
    snippetcompiler_global.tear_down_func()


class CLI(object):
    async def run(self, *args):
        # set column width very wide so lines are not wrapped
        os.environ["COLUMNS"] = "1000"
        runner = testing.CliRunner()
        cmd_args = ["--host", "localhost", "--port", config.Config.get("cmdline_rest_transport", "port")]
        cmd_args.extend(args)

        def invoke():
            return runner.invoke(
                cli=inmanta.main.cmd,
                args=cmd_args,
                catch_exceptions=False
            )

        result = await asyncio.get_event_loop().run_in_executor(None, invoke)
        # reset to default again
        del os.environ["COLUMNS"]
        return result


@pytest.fixture
def cli():
    o = CLI()
    yield o


@pytest.fixture
def postgres_proc(free_port):
    proc = PostgresProc(int(free_port))
    yield proc
    proc.stop()


class AsyncCleaner(object):

    def __init__(self):
        self.register = []

    def add(self, method):
        self.register.append(method)

    def __call__(self, method):
        self.add(method)


@pytest.fixture
async def async_finalizer():
    cleaner = AsyncCleaner()
    yield cleaner
    await asyncio.gather(*[item() for item in cleaner.register])



from inmanta.config import Config
import inmanta.agent.config as cfg


def test_environment_deprecated_options(caplog):
    for (deprecated_option, new_option) in [(cfg.agent_interval, cfg.agent_deploy_interval),
                                            (cfg.agent_splay, cfg.agent_deploy_splay_time)]:

        Config.set(deprecated_option.section, deprecated_option.name, "22")
        caplog.clear()
        assert new_option.get() == 22
        assert "Config option %s is deprecated. Use %s instead." % (deprecated_option.name, new_option.name) in caplog.text

        Config.set(new_option.section, new_option.name, "23")
        caplog.clear()
        assert new_option.get() == 23
        assert "Config option %s is deprecated. Use %s instead." % (deprecated_option.name, new_option.name) not in caplog.text

        Config.load_config()  # Reset config options to default values
        assert new_option.get() != 23
        assert deprecated_option.get() != 23
        Config.set(new_option.section, new_option.name, "24")
        caplog.clear()
        assert new_option.get() == 24
        assert "Config option %s is deprecated. Use %s instead." % (deprecated_option.name, new_option.name) not in caplog.text



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging
import socket
import uuid
from collections import defaultdict

from urllib import parse
from asyncio import Future, ensure_future
from typing import Any, Dict, List, Optional, Union, Tuple, Set, Callable, Generator  # noqa: F401

from inmanta import config as inmanta_config
from inmanta import util
from inmanta.protocol.common import UrlMethod
from inmanta.types import NoneGen
from . import common
from .rest import client

from tornado import ioloop, gen

LOGGER: logging.Logger = logging.getLogger(__name__)


class CallTarget(object):
    """
        A baseclass for all classes that are target for protocol calls / methods
    """

    def _get_endpoint_metadata(self) -> Dict[str, Tuple[str, Callable]]:
        total_dict = {
            method_name: getattr(self, method_name) for method_name in dir(self) if callable(getattr(self, method_name))
        }

        methods: Dict[str, Tuple[str, Callable]] = {}
        for name, attr in total_dict.items():
            if name[0:2] != "__" and hasattr(attr, "__protocol_method__"):
                if attr.__protocol_method__ in methods:
                    raise Exception("Unable to register multiple handlers for the same method. %s" % attr.__protocol_method__)

                methods[attr.__protocol_method__] = (name, attr)

        return methods

    def get_op_mapping(self) -> Dict[str, Dict[str, UrlMethod]]:
        """
            Build a mapping between urls, ops and methods
        """
        url_map: Dict[str, Dict[str, UrlMethod]] = defaultdict(dict)

        # TODO: avoid colliding handlers
        for method, method_handlers in self._get_endpoint_metadata().items():
            properties = method.__method_properties__
            # self.headers.update(properties.get_call_headers())
            url = properties.get_listen_url()
            url_map[url][properties.operation] = UrlMethod(properties, self, method_handlers[1], method_handlers[0])

        return url_map


class Endpoint(object):
    """
        An end-point in the rpc framework
    """

    def __init__(self, name: str):
        self._name: str = name
        self._node_name: str = inmanta_config.nodename.get()
        self._end_point_names: List[str] = []
        self._targets: List[CallTarget] = []

    def add_call_target(self, target: CallTarget) -> None:
        self._targets.append(target)

    @property
    def call_targets(self) -> List[CallTarget]:
        return self._targets

    def add_future(self, future: Future) -> None:
        """
            Add a future to the ioloop to be handled, but do not require the result.
        """

        def handle_result(f: Future) -> None:
            try:
                f.result()
            except Exception as e:
                LOGGER.exception("An exception occurred while handling a future: %s", str(e))

        ioloop.IOLoop.current().add_future(ensure_future(future), handle_result)

    def get_end_point_names(self) -> List[str]:
        return self._end_point_names

    def add_end_point_name(self, name: str) -> None:
        """
            Add an additional name to this endpoint to which it reacts and sends out in heartbeats
        """
        LOGGER.debug("Adding '%s' as endpoint", name)
        self._end_point_names.append(name)

    def clear_end_points(self) -> None:
        """
            Clear all endpoints
        """
        self._end_point_names = []

    name = property(lambda self: self._name)
    end_point_names = property(get_end_point_names)

    def _get_hostname(self) -> str:
        """
            Determine the hostname of this machine
        """
        return socket.gethostname()

    def get_node_name(self) -> str:
        return self._node_name

    node_name = property(get_node_name)


class SessionEndpoint(Endpoint, CallTarget):
    """
        An endpoint for clients that make calls to a server and that receive calls back from the server using long-poll
    """

    def __init__(self, name: str, timeout: int = 120, reconnect_delay: int = 5):
        super().__init__(name)
        self._transport = client.RESTClient
        self._client: Optional[SessionClient] = None
        self._sched = util.Scheduler("session endpoint")

        self._env_id: Optional[uuid.UUID] = None

        self.sessionid: uuid.UUID = uuid.uuid1()
        self.running: bool = True
        self.server_timeout = timeout
        self.reconnect_delay = reconnect_delay

        self.add_call_target(self)

    def get_environment(self) -> Optional[uuid.UUID]:
        return self._env_id

    environment = property(get_environment)

    def set_environment(self, environment_id: uuid.UUID) -> None:
        """
            Set the environment of this agent
        """
        if isinstance(environment_id, str):
            self._env_id = uuid.UUID(environment_id)
        else:
            self._env_id = environment_id

    @gen.coroutine
    def start(self) -> NoneGen:
        """
            Connect to the server and use a heartbeat and long-poll for two-way communication
        """
        assert self._env_id is not None
        LOGGER.log(3, "Starting agent for %s", str(self.sessionid))
        self._client = SessionClient(self.name, self.sessionid, timeout=self.server_timeout)
        ioloop.IOLoop.current().add_callback(self.perform_heartbeat)

    @gen.coroutine
    def stop(self) -> NoneGen:
        self._sched.stop()
        self.running = False

    @gen.coroutine
    def on_reconnect(self) -> NoneGen:
        pass

    @gen.coroutine
    def on_disconnect(self) -> NoneGen:
        pass

    @gen.coroutine
    def perform_heartbeat(self) -> NoneGen:
        """
            Start a continuous heartbeat call
        """
        if self._client is None:
            raise Exception("AgentEndpoint not started")

        connected: bool = False
        while self.running:
            LOGGER.log(3, "sending heartbeat for %s", str(self.sessionid))
            result = yield self._client.heartbeat(
                sid=str(self.sessionid), tid=str(self._env_id), endpoint_names=self.end_point_names, nodename=self.node_name
            )
            LOGGER.log(3, "returned heartbeat for %s", str(self.sessionid))
            if result.code == 200:
                if not connected:
                    connected = True
                    self.add_future(self.on_reconnect())
                if result.result is not None:
                    if "method_calls" in result.result:
                        method_calls: List[common.Request] = [
                            common.Request.from_dict(req) for req in result.result["method_calls"]
                        ]
                        # FIXME: reuse transport?
                        transport = self._transport(self)

                        for method_call in method_calls:
                            self.dispatch_method(transport, method_call)
            else:
                LOGGER.warning(
                    "Heartbeat failed with status %d and message: %s, going to sleep for %d s",
                    result.code,
                    result.result,
                    self.reconnect_delay,
                )
                connected = False
                yield self.on_disconnect()
                yield gen.sleep(self.reconnect_delay)

    def dispatch_method(self, transport: client.RESTClient, method_call: common.Request) -> None:
        if self._client is None:
            raise Exception("AgentEndpoint not started")

        LOGGER.debug("Received call through heartbeat: %s %s %s", method_call.reply_id, method_call.method, method_call.url)
        kwargs, config = transport.match_call(method_call.url, method_call.method)

        if config is None:
            msg = "An error occurred during heartbeat method call (%s %s %s): %s" % (
                method_call.reply_id,
                method_call.method,
                method_call.url,
                "No such method",
            )
            LOGGER.error(msg)
            self.add_future(self._client.heartbeat_reply(self.sessionid, method_call.reply_id, {"result": msg, "code": 500}))

        body = method_call.body or {}
        query_string = parse.urlparse(method_call.url).query
        for key, value in parse.parse_qs(query_string, keep_blank_values=True):
            if len(value) == 1:
                body[key] = value[0].decode("latin-1")
            else:
                body[key] = [v.decode("latin-1") for v in value]

        # FIXME: why create a new transport instance on each call? keep-alive?
        call_result: common.Response = transport._execute_call(kwargs, method_call.method, config, body, method_call.headers)

        def submit_result(future: Future) -> None:
            if future is None:
                return

            response: common.Response = future.result()
            if response.status_code == 500:
                msg = ""
                if response.body is not None and "message" in response.body:
                    msg = response.body["message"]
                LOGGER.error(
                    "An error occurred during heartbeat method call (%s %s %s): %s",
                    method_call.reply_id,
                    method_call.method,
                    method_call.url,
                    msg,
                )

            if self._client is None:
                raise Exception("AgentEndpoint not started")

            self._client.heartbeat_reply(
                self.sessionid, method_call.reply_id, {"result": response.body, "code": response.status_code}
            )

        ioloop.IOLoop.current().add_future(call_result, submit_result)


class Client(Endpoint):
    """
        A client that communicates with end-point based on its configuration
    """

    def __init__(self, name: str, timeout: int = 120) -> None:
        super().__init__(name)
        assert isinstance(timeout, int), "Timeout needs to be an integer value."
        LOGGER.debug("Start transport for client %s", self.name)
        self._transport_instance = client.RESTClient(self, connection_timout=timeout)

    @gen.coroutine
    def _call(self, method_properties: common.MethodProperties, args: List, kwargs: Dict) -> common.Result:
        """
            Execute a call and return the result
        """
        result = yield self._transport_instance.call(method_properties, args, kwargs)
        return result

    def __getattr__(self, name: str) -> Callable:
        """
            Return a function that will call self._call with the correct method properties associated
        """
        if name in common.MethodProperties.methods:
            method = common.MethodProperties.methods[name]

            def wrap(*args: List, **kwargs: Dict) -> common.Result:
                method.function(*args, **kwargs)
                return self._call(method_properties=method, args=args, kwargs=kwargs)

            return wrap

        raise AttributeError("Method with name %s is not defined for this client" % name)


class SyncClient(object):
    """
        A synchronous client that communicates with end-point based on its configuration
    """

    def __init__(self, name: str, timeout: int = 120) -> None:
        self.name = name
        self.timeout = timeout
        self._client = Client(self.name, self.timeout)

    def __getattr__(self, name: str) -> Callable:
        def async_call(*args: List, **kwargs: Dict) -> None:
            method = getattr(self._client, name)

            def method_call() -> None:
                return method(*args, **kwargs)

            try:
                return ioloop.IOLoop.current().run_sync(method_call, self.timeout)
            except TimeoutError:
                raise ConnectionRefusedError()

        return async_call


class SessionClient(Client):
    """
        A client that communicates with server endpoints over a session.
    """
    def __init__(self, name: str, sid: uuid.UUID, timeout: int = 120) -> None:
        super().__init__(name, timeout)
        self._sid = sid

    @gen.coroutine
    def _call(self, method_properties: common.MethodProperties, args: List, kwargs: Dict) -> common.Result:
        """
            Execute the rpc call
        """
        if "sid" not in kwargs:
            kwargs["sid"] = self._sid

        result = yield self._transport_instance.call(method_properties, args, kwargs)
        return result



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from asyncio import Future
from . import common

from typing import Any, Dict, List, Optional, Tuple, Set, Callable, Generator  # noqa: F401


class handle(object):  # noqa: N801
    """
        Decorator for subclasses of an endpoint to handle protocol methods

        :param method A subclass of method that defines the method
        :param kwargs: Map arguments in the message from one name to an other
    """

    def __init__(self, method: Callable[..., Any], **kwargs: str) -> None:
        self.method = method
        self.mapping: Dict[str, str] = kwargs

    def __call__(self, function: Callable[..., Future]):
        """
            The wrapping
        """
        function.__protocol_method__ = self.method
        function.__protocol_mapping__ = self.mapping
        return function


def method(
    method_name: str,
    index: bool = False,
    id: bool = False,
    operation: str = "POST",
    reply: bool = True,
    arg_options: Dict[str, common.ArgOption] = {},
    timeout: Optional[int] = None,
    server_agent: bool = False,
    api: bool = None,
    agent_server: bool = False,
    validate_sid: bool = None,
    client_types: List[str] = ["public"],
    api_version: int = 1,
) -> Callable[..., Callable]:
    """
        Decorator to identify a method as a RPC call. The arguments of the decorator are used by each transport to build
        and model the protocol.

        :param index: A method that returns a list of resources. The url of this method is only the method/resource name.
        :param id: This method requires an id of a resource. The python function should have an id parameter.
        :param operation: The type of HTTP operation (verb)
        :param timeout: nr of seconds before request it terminated
        :param api This is a call from the client to the Server (True if not server_agent and not agent_server)
        :param server_agent: This is a call from the Server to the Agent (reverse http channel through long poll)
        :param agent_server: This is a call from the Agent to the Server
        :param validate_sid: This call requires a valid session, true by default if agent_server and not api
        :param client_types: The allowed client types for this call
        :param arg_options Options related to arguments passed to the method. The key of this dict is the name of the arg to
            which the options apply. The value is another dict that can contain the following options:
                header: Map this argument to a header with the following name.
                reply_header: If the argument is mapped to a header, this header will also be included in the reply
                getter: Call this method after validation and pass its return value to the method call. This may change the
                        type of the argument. This method can raise an HTTPException to return a 404 for example.
    """

    def wrapper(func: Callable[..., Dict[str, Any]]) -> Callable[..., Dict[str, Any]]:
        common.MethodProperties(
            func,
            method_name,
            index,
            id,
            operation,
            reply,
            arg_options,
            timeout,
            server_agent,
            api,
            agent_server,
            validate_sid,
            client_types,
            api_version,
        )
        return func

    return wrapper
